import collections.abc
import os
import warnings

import h5py
import numpy

from ..adapters.utils import IndexersMixin
from ..iterviews import ItemsView, KeysView, ValuesView
from ..structures.core import StructureFamily
from ..utils import node_repr, path_from_uri
from .array import ArrayAdapter

SWMR_DEFAULT = bool(int(os.getenv("TILED_HDF5_SWMR_DEFAULT", "0")))
INLINED_DEPTH = int(os.getenv("TILED_HDF5_INLINED_CONTENTS_MAX_DEPTH", "7"))


def from_dataset(dataset):
    return ArrayAdapter.from_array(dataset, metadata=getattr(dataset, "attrs", {}))


class HDF5Adapter(collections.abc.Mapping, IndexersMixin):
    """
    Read an HDF5 file or a group within one.

    This map the structure of an HDF5 file onto a "Tree" of array structures.

    Examples
    --------

    From the root node of a file given a filepath

    >>> import h5py
    >>> HDF5Adapter.from_uri("file://localhost/path/to/file.h5")

    From the root node of a file given an h5py.File object

    >>> import h5py
    >>> file = h5py.File("path/to/file.h5")
    >>> HDF5Adapter.from_file(file)

    From a group within a file

    >>> import h5py
    >>> file = h5py.File("path/to/file.h5")
    >>> HDF5Adapter(file["some_group']["some_sub_group"])

    """

    structure_family = StructureFamily.container

    def __init__(
        self, node, *, structure=None, metadata=None, specs=None, access_policy=None
    ):
        self._node = node
        self._access_policy = access_policy
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        super().__init__()

    @classmethod
    def from_file(
        cls,
        file,
        *,
        structure=None,
        metadata=None,
        swmr=SWMR_DEFAULT,
        libver="latest",
        specs=None,
        access_policy=None,
    ):
        return cls(file, metadata=metadata, specs=specs, access_policy=access_policy)

    @classmethod
    def from_uri(
        cls,
        data_uri,
        *,
        structure=None,
        metadata=None,
        swmr=SWMR_DEFAULT,
        libver="latest",
        specs=None,
        access_policy=None,
    ):
        filepath = path_from_uri(data_uri)
        file = h5py.File(filepath, "r", swmr=swmr, libver=libver)
        return cls.from_file(file)

    def __repr__(self):
        return node_repr(self, list(self))

    @property
    def access_policy(self):
        return self._access_policy

    def structure(self):
        return None

    def metadata(self):
        d = dict(self._node.attrs)
        for k, v in list(d.items()):
            # Convert any bytes to str.
            if isinstance(v, bytes):
                d[k] = v.decode()
        d.update(self._provided_metadata)
        return d

    def __iter__(self):
        yield from self._node

    def __getitem__(self, key):
        value = self._node[key]
        if isinstance(value, h5py.Group):
            return HDF5Adapter(value)
        else:
            if value.dtype == numpy.dtype("O"):
                warnings.warn(
                    f"The dataset {key} is of object type, using a "
                    "Python-only feature of h5py that is not supported by "
                    "HDF5 in general. Read more about that feature at "
                    "https://docs.h5py.org/en/stable/special.html. "
                    "Consider using a fixed-length field instead. "
                    "Tiled will serve an empty placeholder, unless the "
                    "object is of size 1, where it will attempt to repackage "
                    "the data into a numpy array."
                )

                check_str_dtype = h5py.check_string_dtype(value.dtype)
                if check_str_dtype.length is None:
                    dataset_names = value.file[self._node.name + "/" + key][...][()]
                    if value.size == 1:
                        arr = numpy.array(dataset_names)
                        return from_dataset(arr)
                return from_dataset(numpy.array([]))
            return from_dataset(value)

    def __len__(self):
        return len(self._node)

    def keys(self):
        return KeysView(lambda: len(self), self._keys_slice)

    def values(self):
        return ValuesView(lambda: len(self), self._items_slice)

    def items(self):
        return ItemsView(lambda: len(self), self._items_slice)

    def search(self, query):
        """
        Return a Tree with a subset of the mapping.
        """
        raise NotImplementedError

    def read(self, fields=None):
        if fields is not None:
            raise NotImplementedError
        return self

    # The following two methods are used by keys(), values(), items().

    def _keys_slice(self, start, stop, direction):
        keys = list(self._node)
        if direction < 0:
            keys = reversed(keys)
        return keys[start:stop]

    def _items_slice(self, start, stop, direction):
        items = [(key, self[key]) for key in list(self)]
        if direction < 0:
            items = reversed(items)
        return items[start:stop]

    def inlined_contents_enabled(self, depth):
        return depth <= INLINED_DEPTH



import importlib
import time
from dataclasses import asdict

from ..structures.core import Spec, StructureFamily
from ..utils import UNCHANGED, DictView, ListView, OneShotCachedMap, safe_json_dump
from .utils import MSGPACK_MIME_TYPE, handle_error


class MetadataRevisions:
    def __init__(self, context, link):
        self._cached_len = None
        self.context = context
        self._link = link

    def __len__(self):
        LENGTH_CACHE_TTL = 1  # second

        now = time.monotonic()
        if self._cached_len is not None:
            length, deadline = self._cached_len
            if now < deadline:
                # Used the cached value and do not make any request.
                return length

        content = handle_error(
            self.context.http_client.get(
                self._link,
                headers={"Accept": MSGPACK_MIME_TYPE},
                params={"page[offset]": 0, "page[limit]": 0},
            )
        ).json()
        length = content["meta"]["count"]
        self._cached_len = (length, now + LENGTH_CACHE_TTL)
        return length

    def __getitem__(self, item_):
        self._cached_len = None

        if isinstance(item_, int):
            offset = item_
            limit = 1

            content = handle_error(
                self.context.http_client.get(
                    self._link,
                    headers={"Accept": MSGPACK_MIME_TYPE},
                    params={"page[offset]": offset, "page[limit]": limit},
                )
            ).json()
            (result,) = content["data"]
            return result

        elif isinstance(item_, slice):
            offset = item_.start
            if offset is None:
                offset = 0
            if item_.stop is None:
                params = f"?page[offset]={offset}"
            else:
                limit = item_.stop - offset
                params = f"?page[offset]={offset}&page[limit]={limit}"

            next_page = self._link + params
            result = []
            while next_page is not None:
                content = handle_error(
                    self.context.http_client.get(
                        next_page,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                    )
                ).json()
                if len(result) == 0:
                    result = content.copy()
                else:
                    result["data"].append(content["data"])
                next_page = content["links"]["next"]

            return result["data"]

    def delete_revision(self, n):
        handle_error(self.context.http_client.delete(self._link, params={"number": n}))


class BaseClient:
    def __init__(self, context, *, item, structure_clients, structure=None):
        self._context = context
        self._item = item
        self._cached_len = None  # a cache just for __len__
        self.structure_clients = structure_clients
        self._metadata_revisions = None
        attributes = self.item["attributes"]
        structure_family = attributes["structure_family"]
        if structure is not None:
            # Allow the caller to optionally hand us a structure that is already
            # parsed from a dict into a structure dataclass.
            self._structure = structure
        elif structure_family == StructureFamily.container:
            self._structure = None
        else:
            structure_type = STRUCTURE_TYPES[attributes["structure_family"]]
            self._structure = structure_type.from_json(attributes["structure"])
        super().__init__()

    def structure(self):
        """
        Return a dataclass describing the structure of the data.
        """
        if getattr(self._structure, "resizable", None):
            # In the future, conditionally fetch updated information.
            raise NotImplementedError(
                "The server has indicated that this has a dynamic, resizable "
                "structure and this version of the Tiled Python client cannot "
                "cope with that."
            )
        return self._structure

    def login(self, username=None, provider=None):
        """
        Depending on the server's authentication method, this will prompt for username/password:

        >>> c.login()
        Username: USERNAME
        Password: <input is hidden>

        or prompt you to open a link in a web browser to login with a third party:

        >>> c.login()
        You have ... minutes visit this URL

        https://...

        and enter the code: XXXX-XXXX
        """
        self.context.login(username=username, provider=provider)

    def logout(self):
        """
        Log out.

        This method is idempotent: if you are already logged out, it will do nothing.
        """
        self.context.logout()

    def __repr__(self):
        return f"<{type(self).__name__}>"

    @property
    def context(self):
        return self._context

    @property
    def item(self):
        "JSON payload describing this item. Mostly for internal use."
        return self._item

    @property
    def metadata(self):
        "Metadata about this data source."
        # Ensure this is immutable (at the top level) to help the user avoid
        # getting the wrong impression that editing this would update anything
        # persistent.
        return DictView(self._item["attributes"]["metadata"])

    @property
    def specs(self):
        "List of specifications describing the structure of the metadata and/or data."
        return ListView([Spec(**spec) for spec in self._item["attributes"]["specs"]])

    @property
    def uri(self):
        "Direct link to this entry"
        return self.item["links"]["self"]

    @property
    def structure_family(self):
        "Quick access to this entry"
        return StructureFamily[self.item["attributes"]["structure_family"]]

    def new_variation(self, structure_clients=UNCHANGED, **kwargs):
        """
        This is intended primarily for internal use and use by subclasses.
        """
        if structure_clients is UNCHANGED:
            structure_clients = self.structure_clients
        return type(self)(
            item=self._item,
            structure_clients=structure_clients,
            **kwargs,
        )

    @property
    def formats(self):
        "List formats that the server can export this data as."
        formats = set()
        for spec in self.item["attributes"]["specs"]:
            formats.update(self.context.server_info["formats"].get(spec, []))
        formats.update(
            self.context.server_info["formats"][
                self.item["attributes"]["structure_family"]
            ]
        )
        return sorted(formats)

    def update_metadata(self, metadata=None, specs=None):
        """
        EXPERIMENTAL: Update metadata.

        This is subject to change or removal without notice

        Parameters
        ----------
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        specs : List[str], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        """

        self._cached_len = None

        if specs is None:
            normalized_specs = None
        else:
            normalized_specs = []
            for spec in specs:
                if isinstance(spec, str):
                    spec = Spec(spec)
                normalized_specs.append(asdict(spec))
        data = {
            "metadata": metadata,
            "specs": normalized_specs,
        }

        content = handle_error(
            self.context.http_client.put(
                self.item["links"]["self"], content=safe_json_dump(data)
            )
        ).json()

        if metadata is not None:
            if "metadata" in content:
                # Metadata was accepted and modified by the specs validator on the server side.
                # It is updated locally using the new version.
                self._item["attributes"]["metadata"] = content["metadata"]
            else:
                # Metadata was accepted as it si by the server.
                # It is updated locally with the version submitted buy the client.
                self._item["attributes"]["metadata"] = metadata

        if specs is not None:
            self._item["attributes"]["specs"] = normalized_specs

    @property
    def metadata_revisions(self):
        if self._metadata_revisions is None:
            link = self.item["links"]["self"].replace("/metadata", "/revisions", 1)
            self._metadata_revisions = MetadataRevisions(self.context, link)

        return self._metadata_revisions


STRUCTURE_TYPES = OneShotCachedMap(
    {
        StructureFamily.array: lambda: importlib.import_module(
            "...structures.array", BaseClient.__module__
        ).ArrayStructure,
        StructureFamily.awkward: lambda: importlib.import_module(
            "...structures.awkward", BaseClient.__module__
        ).AwkwardStructure,
        StructureFamily.table: lambda: importlib.import_module(
            "...structures.table", BaseClient.__module__
        ).TableStructure,
        StructureFamily.sparse: lambda: importlib.import_module(
            "...structures.sparse", BaseClient.__module__
        ).SparseStructure,
    }
)



import collections
import os
import secrets
from datetime import timedelta
from functools import lru_cache
from typing import Any, List, Optional

from pydantic import BaseSettings

DatabaseSettings = collections.namedtuple(
    "DatabaseSettings", "uri pool_size pool_pre_ping max_overflow"
)


class Settings(BaseSettings):
    tree: Any = None
    allow_anonymous_access: bool = bool(
        int(os.getenv("TILED_ALLOW_ANONYMOUS_ACCESS", False))
    )
    allow_origins: List[str] = [
        item for item in os.getenv("TILED_ALLOW_ORIGINS", "").split() if item
    ]
    object_cache_available_bytes = float(
        os.getenv("TILED_OBJECT_CACHE_AVAILABLE_BYTES", "0.15")
    )
    object_cache_log_level = os.getenv("TILED_OBJECT_CACHE_LOG_LEVEL", "INFO")
    authenticator: Any = None
    # These 'single user' settings are only applicable if authenticator is None.
    single_user_api_key = os.getenv("TILED_SINGLE_USER_API_KEY", secrets.token_hex(32))
    single_user_api_key_generated = not ("TILED_SINGLE_USER_API_KEY" in os.environ)
    # The TILED_SERVER_SECRET_KEYS may be a single key or a ;-separated list of
    # keys to support key rotation. The first key will be used for encryption. Each
    # key will be tried in turn for decryption.
    secret_keys: List[str] = os.getenv(
        "TILED_SERVER_SECRET_KEYS", secrets.token_hex(32)
    ).split(";")
    access_token_max_age: timedelta = timedelta(
        seconds=int(os.getenv("TILED_ACCESS_TOKEN_MAX_AGE", 15 * 60))  # 15 minutes
    )
    refresh_token_max_age: timedelta = timedelta(
        seconds=int(
            os.getenv("TILED_REFRESH_TOKEN_MAX_AGE", 7 * 24 * 60 * 60)
        )  # 7 days
    )
    session_max_age: Optional[timedelta] = timedelta(
        seconds=int(os.getenv("TILED_SESSION_MAX_AGE", 365 * 24 * 60 * 60))  # 365 days
    )
    # Put a fairly low limit on the maximum size of one chunk, keeping in mind
    # that data should generally be chunked. When we implement async responses,
    # we can raise this global limit.
    response_bytesize_limit = int(
        os.getenv("TILED_RESPONSE_BYTESIZE_LIMIT", 300_000_000)
    )  # 300 MB
    reject_undeclared_specs = bool(int(os.getenv("TILED_REJECT_UNDECLARED_SPECS", 0)))
    database_uri: Optional[str] = os.getenv("TILED_DATABASE_URI")
    database_pool_size: Optional[int] = int(os.getenv("TILED_DATABASE_POOL_SIZE", 5))
    database_pool_pre_ping: Optional[bool] = bool(
        int(os.getenv("TILED_DATABASE_POOL_PRE_PING", 1))
    )
    database_max_overflow: Optional[int] = int(
        os.getenv("TILED_DATABASE_MAX_OVERFLOW", 5)
    )

    @property
    def database_settings(self):
        # The point of this alias is to return a hashable cache key for use in
        # the module tiled.authn_database.connection_pool.
        return DatabaseSettings(
            uri=self.database_uri,
            pool_size=self.database_pool_size,
            pool_pre_ping=self.database_pool_pre_ping,
            max_overflow=self.database_max_overflow,
        )


@lru_cache()
def get_settings():
    return Settings()



import contextlib
import getpass
import uuid

import httpx
import pytest
from sqlalchemy import text
from sqlalchemy.ext.asyncio import create_async_engine

from ..client import context


@contextlib.contextmanager
def fail_with_status_code(status_code):
    with pytest.raises(httpx.HTTPStatusError) as info:
        yield info
    assert info.value.response.status_code == status_code


@contextlib.asynccontextmanager
async def temp_postgres(uri):
    if uri.endswith("/"):
        uri = uri[:-1]
    # Create a fresh database.
    engine = create_async_engine(uri)
    database_name = f"tiled_test_disposable_{uuid.uuid4().hex}"
    async with engine.connect() as connection:
        await connection.execute(
            text("COMMIT")
        )  # close the automatically-started transaction
        await connection.execute(text(f"CREATE DATABASE {database_name};"))
        await connection.commit()
    yield f"{uri}/{database_name}"
    # Drop the database.
    async with engine.connect() as connection:
        await connection.execute(
            text("COMMIT")
        )  # close the automatically-started transaction
        await connection.execute(text(f"DROP DATABASE {database_name};"))
        await connection.commit()


@contextlib.contextmanager
def enter_password(password):
    """
    Override getpass, used like:

    >>> with enter_password(...):
    ...     # Run code that calls getpass.getpass().
    """

    original_prompt = context.PROMPT_FOR_REAUTHENTICATION
    original_getpass = getpass.getpass
    context.PROMPT_FOR_REAUTHENTICATION = True
    setattr(getpass, "getpass", lambda: password)
    yield
    setattr(getpass, "getpass", original_getpass)
    context.PROMPT_FOR_REAUTHENTICATION = original_prompt



import asyncio
import base64
import functools
import logging
import re
import secrets
from collections.abc import Iterable

import httpx
from fastapi import APIRouter, Request
from jose import JWTError, jwk, jwt
from starlette.responses import RedirectResponse

from .server.authentication import Mode
from .server.protocols import UserSessionState
from .server.utils import get_root_url
from .utils import modules_available

logger = logging.getLogger(__name__)


class DummyAuthenticator:
    """
    For test and demo purposes only!

    Accept any username and any password.

    """

    mode = Mode.password

    def __init__(self, confirmation_message=""):
        self.confirmation_message = confirmation_message

    async def authenticate(self, username: str, password: str) -> UserSessionState:
        return UserSessionState(username, {})


class DictionaryAuthenticator:
    """
    For test and demo purposes only!

    Check passwords from a dictionary of usernames mapped to passwords.
    """

    mode = Mode.password
    configuration_schema = """
$schema": http://json-schema.org/draft-07/schema#
type: object
additionalProperties: false
properties:
  users_to_password:
    type: object
    description: |
      Mapping usernames to password. Environment variable expansion should be
      used to avoid placing passwords directly in configuration.
  confirmation_message:
    type: string
    description: May be displayed by client after successful login.
"""

    def __init__(self, users_to_passwords, confirmation_message=""):
        self._users_to_passwords = users_to_passwords
        self.confirmation_message = confirmation_message

    async def authenticate(self, username: str, password: str) -> UserSessionState:
        true_password = self._users_to_passwords.get(username)
        if not true_password:
            # Username is not valid.
            return
        if secrets.compare_digest(true_password, password):
            return UserSessionState(username, {})


class PAMAuthenticator:
    mode = Mode.password
    configuration_schema = """
$schema": http://json-schema.org/draft-07/schema#
type: object
additionalProperties: false
properties:
  service:
    type: string
    description: PAM service. Default is 'login'.
  confirmation_message:
    type: string
    description: May be displayed by client after successful login.
"""

    def __init__(self, service="login", confirmation_message=""):
        if not modules_available("pamela"):
            raise ModuleNotFoundError(
                "This PAMAuthenticator requires the module 'pamela' to be installed."
            )
        self.service = service
        self.confirmation_message = confirmation_message
        # TODO Try to open a PAM session.

    async def authenticate(self, username: str, password: str) -> UserSessionState:
        import pamela

        try:
            pamela.authenticate(username, password, service=self.service)
        except pamela.PAMError:
            # Authentication failed.
            return
        else:
            return UserSessionState(username, {})


class OIDCAuthenticator:
    mode = Mode.external
    configuration_schema = """
$schema": http://json-schema.org/draft-07/schema#
type: object
additionalProperties: false
properties:
  client_id:
    type: string
  client_secret:
    type: string
  token_uri:
    type: string
  authorization_endpoint:
    type: string
  public_keys:
    type: array
    item:
      type: object
      properties:
        - alg:
            type: string
        - e
            type: string
        - kid
            type: string
        - kty
            type: string
        - n
            type: string
        - use
            type: string
      required:
        - alg
        - e
        - kid
        - kty
        - n
        - use
  confirmation_message:
    type: string
    description: May be displayed by client after successful login.
"""

    def __init__(
        self,
        client_id,
        client_secret,
        public_keys,
        token_uri,
        authorization_endpoint,
        confirmation_message,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.confirmation_message = confirmation_message
        self.public_keys = public_keys
        self.token_uri = token_uri
        self.authorization_endpoint = httpx.URL(authorization_endpoint)

    async def authenticate(self, request) -> UserSessionState:
        code = request.query_params["code"]
        # A proxy in the middle may make the request into something like
        # 'http://localhost:8000/...' so we fix the first part but keep
        # the original URI path.
        redirect_uri = f"{get_root_url(request)}{request.url.path}"
        response = await exchange_code(
            self.token_uri, code, self.client_id, self.client_secret, redirect_uri
        )
        response_body = response.json()
        if response.is_error:
            logger.error("Authentication error: %r", response_body)
            return None
        response_body = response.json()
        id_token = response_body["id_token"]
        access_token = response_body["access_token"]
        # Match the kid in id_token to a key in the list of public_keys.
        key = find_key(id_token, self.public_keys)
        try:
            verified_body = jwt.decode(
                id_token, key, access_token=access_token, audience=self.client_id
            )
        except JWTError:
            logger.exception(
                "Authentication error. Unverified token: %r",
                jwt.get_unverified_claims(id_token),
            )
            return None
        return UserSessionState(verified_body["sub"], {})


class KeyNotFoundError(Exception):
    pass


def find_key(token, keys):
    """
    Find a key from the configured keys based on the kid claim of the token

    Parameters
    ----------
    token : token to search for the kid from
    keys:  list of keys

    Raises
    ------
    KeyNotFoundError:
        returned if the token does not have a kid claim

    Returns
    ------
    key: found key object
    """

    unverified = jwt.get_unverified_header(token)
    kid = unverified.get("kid")
    if not kid:
        raise KeyNotFoundError("No 'kid' in token")

    for key in keys:
        if key["kid"] == kid:
            return jwk.construct(key)
    return KeyNotFoundError(
        f"Token specifies {kid} but we have {[k['kid'] for k in keys]}"
    )


async def exchange_code(token_uri, auth_code, client_id, client_secret, redirect_uri):
    """Method that talks to an IdP to exchange a code for an access_token and/or id_token
    Args:
        token_url ([type]): [description]
        auth_code ([type]): [description]
    """
    auth_value = base64.b64encode(f"{client_id}:{client_secret}".encode()).decode()
    response = httpx.post(
        url=token_uri,
        data={
            "grant_type": "authorization_code",
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "code": auth_code,
            "client_secret": client_secret,
        },
        headers={"Authorization": f"Basic {auth_value}"},
    )
    return response


class SAMLAuthenticator:
    mode = Mode.external

    def __init__(
        self,
        saml_settings,  # See EXAMPLE_SAML_SETTINGS below.
        attribute_name,  # which SAML attribute to use as 'id' for Idenity
        confirmation_message="",
    ):
        self.saml_settings = saml_settings
        self.attribute_name = attribute_name
        self.confirmation_message = confirmation_message
        self.authorization_endpoint = "/login"

        router = APIRouter()

        if not modules_available("onelogin"):
            # The PyPI package name is 'python3-saml'
            # but it imports as 'onelogin'.
            # https://github.com/onelogin/python3-saml
            raise ModuleNotFoundError(
                "This SAMLAuthenticator requires 'python3-saml' to be installed."
            )

        from onelogin.saml2.auth import OneLogin_Saml2_Auth

        @router.get("/login")
        async def saml_login(request: Request):
            req = await prepare_saml_from_fastapi_request(request)
            auth = OneLogin_Saml2_Auth(req, self.saml_settings)
            # saml_settings = auth.get_settings()
            # metadata = saml_settings.get_sp_metadata()
            # errors = saml_settings.validate_metadata(metadata)
            # if len(errors) == 0:
            #   print(metadata)
            # else:
            #   print("Error found on Metadata: %s" % (', '.join(errors)))
            callback_url = auth.login()
            response = RedirectResponse(url=callback_url)
            return response

        self.include_routers = [router]

    async def authenticate(self, request) -> UserSessionState:
        if not modules_available("onelogin"):
            raise ModuleNotFoundError(
                "This SAMLAuthenticator requires the module 'oneline' to be installed."
            )
        from onelogin.saml2.auth import OneLogin_Saml2_Auth

        req = await prepare_saml_from_fastapi_request(request, True)
        auth = OneLogin_Saml2_Auth(req, self.saml_settings)
        auth.process_response()  # Process IdP response
        errors = auth.get_errors()  # This method receives an array with the errors
        if errors:
            raise Exception(
                "Error when processing SAML Response: %s %s"
                % (", ".join(errors), auth.get_last_error_reason())
            )
        if auth.is_authenticated():
            # Return a string that the Identity can use as id.
            attribute_as_list = auth.get_attributes()[self.attribute_name]
            # Confused in what situation this would have more than one item....
            assert len(attribute_as_list) == 1
            return UserSessionState(attribute_as_list[0], {})
        else:
            return None


async def prepare_saml_from_fastapi_request(request, debug=False):
    form_data = await request.form()
    rv = {
        "http_host": request.client.host,
        "server_port": request.url.port,
        "script_name": request.url.path,
        "post_data": {},
        "get_data": {}
        # Advanced request options
        # "https": "",
        # "request_uri": "",
        # "query_string": "",
        # "validate_signature_from_qs": False,
        # "lowercase_urlencoding": False
    }
    if request.query_params:
        rv["get_data"] = (request.query_params,)
    if "SAMLResponse" in form_data:
        SAMLResponse = form_data["SAMLResponse"]
        rv["post_data"]["SAMLResponse"] = SAMLResponse
    if "RelayState" in form_data:
        RelayState = form_data["RelayState"]
        rv["post_data"]["RelayState"] = RelayState
    return rv


class LDAPAuthenticator:
    """
    The authenticator code is based on https://github.com/jupyterhub/ldapauthenticator
    The parameter ``use_tls`` was added for convenience of testing.

    Parameters
    ----------
    server_address: str or list(str)
        Address(es) of the LDAP server(s) to contact. A string value may represent a single
        server, a list of strings may represent one or more servers. If a server address
        includes port, then the value of ``server_port`` is ignored, otherwise ``server_port``
        or the default port is used to access the server.

        Could be an IP address or hostname.
    server_port: int or None
        Port on which to contact the LDAP server. Default port is used if ``None``.

        Defaults to ``636`` if ``use_ssl`` is set, ``389`` otherwise.
    use_ssl: boolean
        Use SSL to communicate with the LDAP server.

        Deprecated in version 3 of LDAP. Your LDAP server must be configured to support this, however.
    use_tls: boolean
        Enable/disable TLS if ``use_ssl`` is False. By default TLS is enabled. It should not be disabled
        in production systems.

    connect_timeout: float
        Timeout used for connecting to the LDAP server. Default: 5.

    receive_timeout: float
        Timeout used for communication with the LDAP server, e.g. this timeout is used to wait for
        completion of 2FA. For smooth operation it should probably exceed timeout set at LDAP server.
        Default: 60.

    bind_dn_template: list or str
        Template from which to construct the full dn
        when authenticating to LDAP. ``{username}`` is replaced
        with the actual username used to log in.

        If your LDAP is set in such a way that the userdn can not
        be formed from a template, but must be looked up with an attribute
        (such as uid or ``sAMAccountName``), please see ``lookup_dn``. It might
        be particularly relevant for ActiveDirectory installs.

        Unicode Example:

        .. code-block::

            "uid={username},ou=people,dc=wikimedia,dc=org"

        List Example:

        .. code-block::

            [
                "uid={username},ou=people,dc=wikimedia,dc=org",
                "uid={username},ou=Developers,dc=wikimedia,dc=org"
                ]
    allowed_groups: list or None
        List of LDAP group DNs that users could be members of to be granted access.

        If a user is in any one of the listed groups, then that user is granted access.
        Membership is tested by fetching info about each group and looking for the User's
        dn to be a value of one of `member` or `uniqueMember`, *or* if the username being
        used to log in with is value of the `uid`.

        Set to an empty list or None to allow all users that have an LDAP account to log in,
        without performing any group membership checks.
    valid_username_regex: str
        Regex for validating usernames - those that do not match this regex will be rejected.

        This is primarily used as a measure against LDAP injection, which has fatal security
        considerations. The default works for most LDAP installations, but some users might need
        to modify it to fit their custom installs. If you are modifying it, be sure to understand
        the implications of allowing additional characters in usernames and what that means for
        LDAP injection issues. See https://www.owasp.org/index.php/LDAP_injection for an overview
        of LDAP injection.
    lookup_dn: boolean
        Form user's DN by looking up an entry from directory

        By default, LDAPAuthenticator finds the user's DN by using `bind_dn_template`.
        However, in some installations, the user's DN does not contain the username, and
        hence needs to be looked up. You can set this to True and then use ``user_search_base``
        and ``user_attribute`` to accomplish this.
    user_search_base: str
        Base for looking up user accounts in the directory, if `lookup_dn` is set to True.

        LDAPAuthenticator will search all objects matching under this base where the `user_attribute`
        is set to the current username to form the userdn.

        For example, if all users objects existed under the base ou=people,dc=wikimedia,dc=org, and
        the username users use is set with the attribute `uid`, you can use the following config:

        .. code-block::

            c.LDAPAuthenticator.lookup_dn = True
            c.LDAPAuthenticator.lookup_dn_search_filter = '({login_attr}={login})'
            c.LDAPAuthenticator.lookup_dn_search_user = 'ldap_search_user_technical_account'
            c.LDAPAuthenticator.lookup_dn_search_password = 'secret'
            c.LDAPAuthenticator.user_search_base = 'ou=people,dc=wikimedia,dc=org'
            c.LDAPAuthenticator.user_attribute = 'sAMAccountName'
            c.LDAPAuthenticator.lookup_dn_user_dn_attribute = 'cn'
            c.LDAPAuthenticator.bind_dn_template = '{username}'
    user_attribute: str
        Attribute containing user's name, if ``lookup_dn`` is set to True.

        See ``user_search_base`` for info on how this attribute is used.

        For most LDAP servers, this is uid.  For Active Directory, it is
        sAMAccountName.
    lookup_dn_search_filter: str or None
        How to query LDAP for user name lookup, if ``lookup_dn`` is set to True.
    lookup_dn_search_user: str or None
        Technical account for user lookup, if ``lookup_dn`` is set to True.

        If both lookup_dn_search_user and lookup_dn_search_password are None,
        then anonymous LDAP query will be done.
    lookup_dn_search_password: str or None
        Technical account for user lookup, if ``lookup_dn`` is set to True.
    lookup_dn_user_dn_attribute: str or None
        Attribute containing user's name needed for  building DN string, if ``lookup_dn`` is set to True.

        See ``user_search_base`` for info on how this attribute is used.

        For most LDAP servers, this is username.  For Active Directory, it is cn.
    escape_userdn: boolean
        If set to True, escape special chars in userdn when authenticating in LDAP.

        On some LDAP servers, when userdn contains chars like '(', ')', '\'
        authentication may fail when those chars
        are not escaped.
    search_filter: str
        LDAP3 Search Filter whose results are allowed access
    attributes: list or None
        List of attributes to be searched
    auth_state_attributes: list or None
        List of attributes to be returned in auth_state for a user
    use_lookup_dn_username: boolean
        If set to true uses the ``lookup_dn_user_dn_attribute`` attribute as username instead of
        the supplied one.

        This can be useful in an heterogeneous environment, when supplying a UNIX username
        to authenticate against AD.
    confirmation_message: str
        May be displayed by client after successful login.

    Examples
    --------

    Using the authenticator class (the code runs in ``asyncio`` loop):

    .. code-block::

        from bluesky_httpserver.authenticators import LDAPAuthenticator
        authenticator = LDAPAuthenticator(
            "localhost", 1389, bind_dn_template="cn={username},ou=users,dc=example,dc=org", use_tls=False
        )
        await authenticator.authenticate("user01", "password1")
        await authenticator.authenticate("user02", "password2")


    Simple example of a config file (e.g. ``config_ldap.yml``):

    .. code-block::

        uvicorn:
            host: localhost
            port: 60610
        authentication:
            providers:
                - provider: ldap_local
                authenticator: bluesky_httpserver.authenticators:LDAPAuthenticator
                args:
                    server_address: localhost
                    server_port: 1389
                    bind_dn_template: "cn={username},ou=users,dc=example,dc=org"
                    use_tls: false
                    use_ssl: false
            tiled_admins:
                - provider: ldap_local
                id: user02
    """

    mode = Mode.password

    def __init__(
        self,
        server_address,
        server_port=None,
        *,
        use_ssl=False,
        use_tls=True,
        connect_timeout=5,
        receive_timeout=60,
        bind_dn_template=None,
        allowed_groups=None,
        valid_username_regex=r"^[a-z][.a-z0-9_-]*$",
        lookup_dn=False,
        user_search_base=None,
        user_attribute=None,
        lookup_dn_search_filter="({login_attr}={login})",
        lookup_dn_search_user=None,
        lookup_dn_search_password=None,
        lookup_dn_user_dn_attribute=None,
        escape_userdn=False,
        search_filter="",
        attributes=None,
        auth_state_attributes=None,
        use_lookup_dn_username=True,
        confirmation_message="",
    ):
        self.use_ssl = use_ssl
        self.use_tls = use_tls
        self.connect_timeout = connect_timeout
        self.receive_timeout = receive_timeout
        self.bind_dn_template = bind_dn_template
        self.allowed_groups = allowed_groups
        self.valid_username_regex = valid_username_regex
        self.lookup_dn = lookup_dn
        self.user_search_base = user_search_base
        self.user_attribute = user_attribute
        self.lookup_dn_search_filter = lookup_dn_search_filter
        self.lookup_dn_search_user = lookup_dn_search_user
        self.lookup_dn_search_password = lookup_dn_search_password
        self.lookup_dn_user_dn_attribute = lookup_dn_user_dn_attribute
        self.escape_userdn = escape_userdn
        self.search_filter = search_filter
        self.attributes = attributes if attributes else []
        self.auth_state_attributes = (
            auth_state_attributes if auth_state_attributes else []
        )
        self.use_lookup_dn_username = use_lookup_dn_username

        if isinstance(server_address, str):
            server_address_list = [server_address]
        elif isinstance(server_address, Iterable):
            server_address_list = list(server_address)
        else:
            raise TypeError(
                f"Unsupported type of `server_address` (list): server_address={server_address} "
                f"type(server_address)={type(server_address)}"
            )
        if not server_address_list:
            raise ValueError(
                "No servers are specified: 'server_address' is an empty list"
            )

        self.server_address_list = server_address_list
        self.server_port = (
            server_port if server_port is not None else self._server_port_default()
        )
        self.confirmation_message = confirmation_message

    def _server_port_default(self):
        if self.use_ssl:
            return 636  # default SSL port for LDAP
        else:
            return 389  # default plaintext port for LDAP

    async def resolve_username(self, username_supplied_by_user):
        import ldap3

        search_dn = self.lookup_dn_search_user
        if self.escape_userdn:
            search_dn = ldap3.utils.conv.escape_filter_chars(search_dn)
        conn = await asyncio.get_running_loop().run_in_executor(
            None, self.get_connection, search_dn, self.lookup_dn_search_password
        )
        is_bound = await asyncio.get_running_loop().run_in_executor(None, conn.bind)
        if not is_bound:
            msg = "Failed to connect to LDAP server with search user '{search_dn}'"
            self.log.warning(msg.format(search_dn=search_dn))
            return (None, None)

        search_filter = self.lookup_dn_search_filter.format(
            login_attr=self.user_attribute, login=username_supplied_by_user
        )
        msg = "\n".join(
            [
                "Looking up user with:",
                "    search_base = '{search_base}'",
                "    search_filter = '{search_filter}'",
                "    attributes = '{attributes}'",
            ]
        )
        logger.debug(
            msg.format(
                search_base=self.user_search_base,
                search_filter=search_filter,
                attributes=self.user_attribute,
            )
        )

        search_func = functools.partial(
            conn.search,
            search_base=self.user_search_base,
            search_scope=ldap3.SUBTREE,
            search_filter=search_filter,
            attributes=[self.lookup_dn_user_dn_attribute],
        )
        await asyncio.get_running_loop().run_in_executor(None, search_func)

        response = conn.response
        if len(response) == 0 or "attributes" not in response[0].keys():
            msg = (
                "No entry found for user '{username}' "
                "when looking up attribute '{attribute}'"
            )
            logger.warning(
                msg.format(
                    username=username_supplied_by_user, attribute=self.user_attribute
                )
            )
            return (None, None)

        user_dn = response[0]["attributes"][self.lookup_dn_user_dn_attribute]
        if isinstance(user_dn, list):
            if len(user_dn) == 0:
                return (None, None)
            elif len(user_dn) == 1:
                user_dn = user_dn[0]
            else:
                msg = (
                    "A lookup of the username '{username}' returned a list "
                    "of entries for the attribute '{attribute}'. Only the "
                    "first among these ('{first_entry}') was used. The other "
                    "entries ({other_entries}) were ignored."
                )
                logger.warning(
                    msg.format(
                        username=username_supplied_by_user,
                        attribute=self.lookup_dn_user_dn_attribute,
                        first_entry=user_dn[0],
                        other_entries=", ".join(user_dn[1:]),
                    )
                )
                user_dn = user_dn[0]

        return (user_dn, response[0]["dn"])

    def get_connection(self, userdn, password):
        import ldap3

        # NOTE: setting 'acitve=False' essentially disables exclusion of inactive servers from the pool.
        # It probably does not matter if the pool contains only one server, but it could have implications
        # when there are multiple servers in the pool. It is not clear what those implications are.
        # But using the default 'activate=True' results in the thread being blocked indefinitely
        # at the step of creating 'ldap3.Connection' regardless of timeouts in case all the servers are
        # inactive (e.g. the pool has one server and it is unaccessible), which is unacceptable.
        # Further investigation may be needed in the future.
        server_pool = ldap3.ServerPool(None, ldap3.RANDOM, active=False)
        for address in self.server_address_list:
            if re.search(r".+:\d+", address):
                # Port is found in the address
                address_split = address.split(":")
                server_addr = ":".join(address_split[:-1])
                server_port = int(address_split[-1])
            else:
                # Use the default port
                server_addr = address
                server_port = self.server_port

            server = ldap3.Server(
                server_addr,
                port=server_port,
                use_ssl=self.use_ssl,
                connect_timeout=self.connect_timeout,
            )
            server_pool.add(server)

        auto_bind_no_ssl = (
            ldap3.AUTO_BIND_TLS_BEFORE_BIND if self.use_tls else ldap3.AUTO_BIND_NO_TLS
        )
        auto_bind = ldap3.AUTO_BIND_NO_TLS if self.use_ssl else auto_bind_no_ssl
        conn = ldap3.Connection(
            server_pool,
            user=userdn,
            password=password,
            auto_bind=auto_bind,
            receive_timeout=self.receive_timeout,
        )
        return conn

    async def get_user_attributes(self, conn, userdn):
        attrs = {}
        if self.auth_state_attributes:
            search_func = functools.partial(
                conn.search,
                userdn,
                "(objectClass=*)",
                attributes=self.auth_state_attributes,
            )
            found = await asyncio.get_running_loop().run_in_executor(None, search_func)
            if found:
                attrs = conn.entries[0].entry_attributes_as_dict
        return attrs

    async def authenticate(self, username: str, password: str) -> UserSessionState:
        import ldap3

        username_saved = username  # Save the user name passed as a parameter

        # Protect against invalid usernames as well as LDAP injection attacks
        if not re.match(self.valid_username_regex, username):
            logger.warning(
                "username:%s Illegal characters in username, must match regex %s",
                username,
                self.valid_username_regex,
            )
            return None

        # No empty passwords!
        if password is None or password.strip() == "":
            logger.warning("username:%s Login denied for blank password", username)
            return None

        # bind_dn_template should be of type List[str]
        bind_dn_template = self.bind_dn_template
        if isinstance(bind_dn_template, str):
            bind_dn_template = [bind_dn_template]

        # sanity check
        if not self.lookup_dn and not bind_dn_template:
            logger.warning(
                "Login not allowed, please configure 'lookup_dn' or 'bind_dn_template'."
            )
            return None

        if self.lookup_dn:
            username, resolved_dn = await self.resolve_username(username)
            if not username:
                return None
            if str(self.lookup_dn_user_dn_attribute).upper() == "CN":
                # Only escape commas if the lookup attribute is CN
                username = re.subn(r"([^\\]),", r"\1\,", username)[0]
            if not bind_dn_template:
                bind_dn_template = [resolved_dn]

        is_bound = False
        for dn in bind_dn_template:
            if not dn:
                logger.warning("Ignoring blank 'bind_dn_template' entry!")
                continue
            userdn = dn.format(username=username)
            if self.escape_userdn:
                userdn = ldap3.utils.conv.escape_filter_chars(userdn)
            msg = "Attempting to bind {username} with {userdn}"
            logger.debug(msg.format(username=username, userdn=userdn))
            msg = "Status of user bind {username} with {userdn} : {is_bound}"
            try:
                conn = await asyncio.get_running_loop().run_in_executor(
                    None, self.get_connection, userdn, password
                )
            except ldap3.core.exceptions.LDAPBindError as exc:
                is_bound = False
                msg += "\n{exc_type}: {exc_msg}".format(
                    exc_type=exc.__class__.__name__,
                    exc_msg=exc.args[0] if exc.args else "",
                )
            else:
                if conn.bound:
                    is_bound = True
                else:
                    is_bound = await asyncio.get_running_loop().run_in_executor(
                        None, conn.bind
                    )

            msg = msg.format(username=username, userdn=userdn, is_bound=is_bound)
            logger.debug(msg)
            if is_bound:
                break

        if not is_bound:
            msg = "Invalid password for user '{username}'"
            logger.warning(msg.format(username=username))
            return None

        if self.search_filter:
            search_filter = self.search_filter.format(
                userattr=self.user_attribute, username=username
            )

            search_func = functools.partial(
                conn.search,
                search_base=self.user_search_base,
                search_scope=ldap3.SUBTREE,
                search_filter=search_filter,
                attributes=self.attributes,
            )
            await asyncio.get_running_loop().run_in_executor(None, search_func)

            n_users = len(conn.response)
            if n_users == 0:
                msg = "User with '{userattr}={username}' not found in directory"
                logger.warning(
                    msg.format(userattr=self.user_attribute, username=username)
                )
                return None
            if n_users > 1:
                msg = (
                    "Duplicate users found! "
                    "{n_users} users found with '{userattr}={username}'"
                )
                logger.warning(
                    msg.format(
                        userattr=self.user_attribute, username=username, n_users=n_users
                    )
                )
                return None

        if self.allowed_groups:
            logger.debug("username:%s Using dn %s", username, userdn)
            found = False
            for group in self.allowed_groups:
                group_filter = (
                    "(|"
                    "(member={userdn})"
                    "(uniqueMember={userdn})"
                    "(memberUid={uid})"
                    ")"
                )
                group_filter = group_filter.format(userdn=userdn, uid=username)
                group_attributes = ["member", "uniqueMember", "memberUid"]

                search_func = functools.partial(
                    conn.search,
                    group,
                    search_scope=ldap3.BASE,
                    search_filter=group_filter,
                    attributes=group_attributes,
                )
                found = await asyncio.get_running_loop().run_in_executor(
                    None, search_func
                )
                if found:
                    break

            if not found:
                # If we reach here, then none of the groups matched
                msg = "username:{username} User not in any of the allowed groups"
                logger.warning(msg.format(username=username))
                return None

        if not self.use_lookup_dn_username:
            username = username_saved

        user_info = await self.get_user_attributes(conn, userdn)
        if user_info:
            logger.debug("username:%s attributes:%s", username, user_info)
            # this path might never have been worked out...is it ever hit?
            return UserSessionState(username, user_info)
        return UserSessionState(username, {})



import base64
import builtins
import collections.abc
import contextlib
import enum
import importlib
import importlib.util
import inspect
import operator
import os
import platform
import sys
import threading
from pathlib import Path
from typing import Any, Callable
from urllib.parse import urlparse

import anyio


class ListView(collections.abc.Sequence):
    "An immutable view of a list."

    def __init__(self, seq):
        self._internal_list = list(seq)

    def __repr__(self):
        return f"{type(self).__name__}({self._internal_list!r})"

    def _repr_pretty_(self, p, cycle):
        """
        Provide "pretty" display in IPython/Jupyter.

        See https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
        """
        from pprint import pformat

        list_ = self._internal_list
        # pformat only works on literal list, so convert if we have
        # some other list-like
        if not type(list_) is list:
            list_ = list(self._internal_list)
        p.text(pformat(list_))

    def __getitem__(self, index):
        return self._internal_list[index]

    def __iter__(self):
        yield from self._internal_list

    def __len__(self):
        return len(self._internal_list)

    def __setitem__(self, index, value):
        raise TypeError("Setting values is not allowed.")

    def __delitem__(self, index):
        raise TypeError("Deleting values is not allowed.")

    def __add__(self, other):
        return type(self)(self._internal_list + other)

    def __eq__(self, other):
        return self._internal_list == other


class DictView(collections.abc.Mapping):
    "An immutable view of a dict."

    def __init__(self, d):
        self._internal_dict = d

    def __repr__(self):
        return f"{type(self).__name__}({self._internal_dict!r})"

    def _repr_pretty_(self, p, cycle):
        """
        Provide "pretty" display in IPython/Jupyter.

        See https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
        """
        from pprint import pformat

        d = self._internal_dict
        # pformat only works on literal dict, so convert if we have
        # some other dict-like
        if not type(d) is dict:
            d = dict(self._internal_dict)
        p.text(pformat(d))

    def __getitem__(self, key):
        return self._internal_dict[key]

    def __iter__(self):
        yield from self._internal_dict

    def __len__(self):
        return len(self._internal_dict)

    def __setitem__(self, key, value):
        raise TypeError("Setting items is not allowed.")

    def __delitem__(self, key):
        raise TypeError("Deleting items is not allowed.")

    def __eq__(self, other):
        return self._internal_dict == other


# This object should never be directly instantiated by external code.
# It is defined at module scope because namedtupled need to be defined at
# module scope in order to be pickleable. It should only be used internally by
# OneShotCachedMap below.
_OneShotCachedMapWrapper = collections.namedtuple("_OneShotCachedMapWrapper", ("func",))


class OneShotCachedMap(collections.abc.Mapping):
    __slots__ = ("__mapping", "__lock")

    def __init__(self, *args, **kwargs):
        dictionary = dict(*args, **kwargs)
        wrap = _OneShotCachedMapWrapper
        # TODO should be recursive lock?
        self.__lock = threading.Lock()
        # TODO type validation?
        self.__mapping = {k: wrap(v) for k, v in dictionary.items()}

    def __getitem__(self, key):
        # TODO per-key locking?
        with self.__lock:
            v = self.__mapping[key]
            if isinstance(v, _OneShotCachedMapWrapper):
                # TODO handle exceptions?
                v = self.__mapping[key] = v.func()
        return v

    def set(self, key, value_factory):
        """
        Set key to a callable the returns value.
        """
        if not callable(value_factory):
            raise ValueError(
                "This requires a callable that return a value, not the value itself."
            )
        self.__mapping[key] = _OneShotCachedMapWrapper(value_factory)

    def discard(self, key):
        """
        Discard a key if it is present. This is idempotent.
        """
        self.__mapping.pop(key, None)
        self.evict(key)

    def remove(self, key):
        """
        Remove a key. Raises KeyError if key not present.
        """
        del self.__mapping[key]
        self.evict(key)

    def __len__(self):
        return len(self.__mapping)

    def __iter__(self):
        return iter(self.__mapping)

    def __contains__(self, k):
        # make sure checking 'in' does not trigger evaluation
        return k in self.__mapping

    def __getstate__(self):
        return self.__mapping

    def __setstate__(self, mapping):
        self.__mapping = mapping
        self.__lock = threading.Lock()

    def __repr__(self):
        d = {}
        for k, v in self.__mapping.items():
            if isinstance(v, _OneShotCachedMapWrapper):
                d[k] = "<lazy>"
            else:
                d[k] = repr(v)
        return (
            f"<{type(self).__name__}"
            "({" + ", ".join(f"{k!r}: {v!s}" for k, v in d.items()) + "})>"
        )


class CachingMap(collections.abc.Mapping):
    """
    Mapping that computes values on read and caches them in a configured cache.

    Parameters
    ----------
    mapping : dict-like
        Must map keys to callables that return values.
    cache : dict-like
        Will be used to cache values. May be ordinary dict, LRUCache, etc.
    """

    __slots__ = ("__mapping", "__cache")

    def __init__(self, mapping, cache):
        self.__mapping = mapping
        self.__cache = cache

    def __getitem__(self, key):
        try:
            return self.__cache[key]
        except KeyError:
            value = self.__mapping[key]()
            self.__cache[key] = value
            return value

    def set(self, key, value_factory):
        """
        Set key to a callable the returns value.
        """
        if not callable(value_factory):
            raise ValueError(
                "This requires a callable that return a value, not the value itself."
            )
        self.__mapping[key] = value_factory
        # This may be replacing (updating) an existing key. Clear any cached value.
        self.evict(key)

    def discard(self, key):
        """
        Discard a key if it is present. This is idempotent.
        """
        self.__mapping.pop(key, None)
        self.evict(key)

    def remove(self, key):
        """
        Remove a key. Raises KeyError if key not present.
        """
        del self.__mapping[key]
        self.evict(key)

    def evict(self, key):
        """
        Evict a key from the internal cache. This is idempotent.

        This does *not* remove the key from the mapping.
        If it is accessed, it will be recomputed and added back to the cache.
        """
        self.__cache.pop(key, None)

    def __len__(self):
        return len(self.__mapping)

    def __iter__(self):
        return iter(self.__mapping)

    def __contains__(self, key):
        # Ensure checking 'in' does not trigger evaluation.
        return key in self.__mapping

    def __getstate__(self):
        return self.__mapping, self.__cache

    def __setstate__(self, state):
        mapping, cache = state
        self.__mapping = mapping
        self.__cache = cache

    def __repr__(self):
        if self.__cache is None:
            d = {k: "<lazy>" for k in self.__mapping}
        else:
            d = {}
            for k in self.__mapping:
                try:
                    value = self.__cache[k]
                except KeyError:
                    d[k] = "<lazy>"
                else:
                    d[k] = repr(value)
        return (
            f"<{type(self).__name__}"
            "({" + ", ".join(f"{k!r}: {v!s}" for k, v in d.items()) + "})>"
        )


class SpecialUsers(str, enum.Enum):
    public = "public"
    admin = "admin"


def _line(nodes, last):
    "Generate a single line for the tree utility"
    tee = "├"
    vertical = "│   "
    horizontal = "── "
    L = "└"
    blank = "    "
    indent = ""
    for item in last[:-1]:
        if item:
            indent += blank
        else:
            indent += vertical
    if last[-1]:
        return indent + L + horizontal + nodes[-1]
    else:
        return indent + tee + horizontal + nodes[-1]


def walk(tree, nodes=None):
    "Walk the entries in a (nested) Tree depth first."
    if nodes is None:
        for node in tree:
            yield from walk(tree, [node])
    else:
        value = tree[nodes[-1]]
        if hasattr(value, "items"):
            yield nodes
            for k, v in value.items():
                yield from walk(value, nodes + [k])
        else:
            yield nodes


def gen_tree(tree, nodes=None, last=None):
    "A generator of lines for the tree utility"

    # Normally, traversing a Tree will cause the structure clients to be
    # instanitated which in turn triggers import of the associated libraries like
    # numpy, pandas, and xarray. We want to avoid paying for that, especially
    # when this function is used in a CLI where import overhead can accumulate to
    # about 2 seconds, the bulk of the time. Therefore, we do something a bit
    # "clever" here to override the normal structure clients with dummy placeholders.
    from .client.container import Container

    def dummy_client(*args, **kwargs):
        return None

    structure_clients = collections.defaultdict(lambda: dummy_client)
    structure_clients["container"] = Container
    fast_tree = tree.new_variation(structure_clients=structure_clients)
    if nodes is None:
        last_index = len(fast_tree) - 1
        for index, node in enumerate(fast_tree):
            yield from gen_tree(fast_tree, [node], [index == last_index])
    else:
        value = fast_tree[nodes[-1]]
        if hasattr(value, "items"):
            yield _line(nodes, last)
            last_index = len(value) - 1
            for index, (k, v) in enumerate(value.items()):
                yield from gen_tree(value, nodes + [k], last + [index == last_index])
        else:
            yield _line(nodes, last)


def tree(tree, max_lines=20):
    """
    Print a visual sketch of Tree structure akin to UNIX `tree`.

    Parameters
    ----------
    tree : Tree
    max_lines: int or None, optional
        By default, output is trucated at 20 lines. ``None`` means "Do not
        truncate."

    Examples
    --------

    >>> tree(tree)
    ├── A
    │   ├── dog
    │   ├── cat
    │   └── monkey
    └── B
        ├── snake
        ├── bear
        └── wolf

    """
    if len(tree) == 0:
        print("<Empty>")
        return
    for counter, line in enumerate(gen_tree(tree), start=1):
        if (max_lines is not None) and (counter > max_lines):
            print(
                f"<Output truncated at {max_lines} lines. "
                "Adjust tree's max_lines parameter to see more.>"
            )
            break
        print(line)


class Sentinel:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"<{self.name}>"

    def __copy__(self):
        # The goal here is to make copy.copy(sentinel) == sentinel
        return self

    def __deepcopy__(self, memo):
        # The goal here is to make copy.deepcopy(sentinel) == sentinel
        return self


UNCHANGED = Sentinel("UNCHANGED")
UNSET = Sentinel("UNSET")


def import_object(colon_separated_string, accept_live_object=True):
    if not isinstance(colon_separated_string, str):
        if not accept_live_object:
            raise ValueError(
                "accept_live_object is False, but live object has been passed"
            )

        # We have been handed the live object itself.
        # Nothing to import. Pass it through.
        return colon_separated_string
    MESSAGE = (
        "Expected string formatted like:\n\n"
        "    package_name.module_name:object_name\n\n"
        "Notice *dots* between modules and a "
        f"*colon* before the object name. Received:\n\n{colon_separated_string!r}"
    )
    import_path, _, obj_path = colon_separated_string.partition(":")
    for segment in import_path.split("."):
        if not segment.isidentifier():
            raise ValueError(MESSAGE)
    for attr in obj_path.split("."):
        if not attr.isidentifier():
            raise ValueError(MESSAGE)
    module = importlib.import_module(import_path)
    return operator.attrgetter(obj_path)(module)


def modules_available(*module_names):
    for module_name in module_names:
        if not importlib.util.find_spec(module_name):
            break
    else:
        # All modules were found.
        return True
    return False


def parse(file):
    """
    Given a config file, parse it.

    This wraps YAML parsing and environment variable expansion.
    """
    import yaml

    content = yaml.safe_load(file.read())
    return expand_environment_variables(content)


def expand_environment_variables(config):
    """Expand environment variables in a nested config dictionary

    VENDORED FROM dask.config.

    This function will recursively search through any nested dictionaries
    and/or lists.

    Parameters
    ----------
    config : dict, iterable, or str
        Input object to search for environment variables

    Returns
    -------
    config : same type as input

    Examples
    --------
    >>> expand_environment_variables({'x': [1, 2, '$USER']})  # doctest: +SKIP
    {'x': [1, 2, 'my-username']}
    """
    if isinstance(config, collections.abc.Mapping):
        return {k: expand_environment_variables(v) for k, v in config.items()}
    elif isinstance(config, str):
        return os.path.expandvars(config)
    elif isinstance(config, (list, tuple, builtins.set)):
        return type(config)([expand_environment_variables(v) for v in config])
    else:
        return config


@contextlib.contextmanager
def prepend_to_sys_path(*paths):
    "Temporarily prepend items to sys.path."

    for item in reversed(paths):
        # Ensure item is str (not pathlib.Path).
        sys.path.insert(0, str(item))
    try:
        yield
    finally:
        for item in paths:
            sys.path.pop(0)


def safe_json_dump(content):
    """
    Baes64-encode raw bytes, and provide a fallback if orjson numpy handling fails.
    """
    import orjson

    def default(content):
        if isinstance(content, bytes):
            content = f"data:application/octet-stream;base64,{base64.b64encode(content).decode('utf-8')}"
            return content
        # No need to import numpy if it hasn't been used already.
        numpy = sys.modules.get("numpy", None)
        if numpy is not None:
            if isinstance(content, numpy.ndarray):
                # If we make it here, OPT_NUMPY_SERIALIZE failed because we have hit some edge case.
                # Give up on the numpy fast-path and convert to Python list.
                # If the items in this list aren't serializable (e.g. bytes) we'll recurse on each item.
                return content.tolist()
            elif isinstance(content, (bytes, numpy.bytes_)):
                return content.decode("utf-8")
        raise TypeError

    # Not all numpy dtypes are supported by orjson.
    # Fall back to converting to a (possibly nested) Python list.
    return orjson.dumps(content, option=orjson.OPT_SERIALIZE_NUMPY, default=default)


class MissingDependency(ModuleNotFoundError):
    pass


class UnrecognizedExtension(ValueError):
    pass


class SerializationError(Exception):
    pass


class UnsupportedShape(SerializationError):
    pass


class UnsupportedQueryType(TypeError):
    pass


class Conflicts(Exception):
    "Prompts the server to send 409 Conflicts with message"
    pass


# Arrow obtained an official MIME type 2021-06-23.
# https://www.iana.org/assignments/media-types/application/vnd.apache.arrow.file
APACHE_ARROW_FILE_MIME_TYPE = "application/vnd.apache.arrow.file"
XLSX_MIME_TYPE = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"

"""Get the data files for this package."""


def get_share_tiled_path():
    """
    Walk up until we find share/tiled.

    Because it is outside the pacakge, its location relative to us is up to the
    package manager. Rather than assuming that it is a specific number of
    directory levels above us, we walk upward until we find it. A file with
    long unique name helps us confirm that we have found the right directory.

    It's not clear whether being this flexible about the location of
    shared/tiled is really necessary, but JupyterHub uses a similar approach,
    so it seems possible that there is variation in the wild and that this is
    the right way to cope with that.
    """
    import sys
    from os.path import abspath, dirname, exists, join, split

    path = abspath(dirname(__file__))
    starting_points = [path]
    if not path.startswith(sys.prefix):
        starting_points.append(sys.prefix)
    for path in starting_points:
        # walk up, looking for prefix/share/tiled
        while path != "/":
            share_tiled = join(path, "share", "tiled")
            if exists(
                join(share_tiled, ".identifying_file_72628d5f953b4229b58c9f1f8f6a9a09")
            ):
                # We have the found the right directory,
                # or one that is trying very hard to pretend to be!
                return share_tiled
            path, _ = split(path)
    # Give up
    return ""


# Package managers can just override this with the appropriate constant
SHARE_TILED_PATH = get_share_tiled_path()


def node_repr(tree, sample):
    sample_reprs = list(map(repr, sample))
    out = f"<{type(tree).__name__} {{"
    # Always show at least one.
    if sample_reprs:
        out += sample_reprs[0]
    # And then show as many more as we can fit on one line.
    counter = 1
    for sample_repr in sample_reprs[1:]:
        if len(out) + len(sample_repr) > 60:  # character count
            break
        out += ", " + sample_repr
        counter += 1
    approx_len = operator.length_hint(tree)  # cheaper to compute than len(tree)
    # Are there more in the tree that what we displayed above?
    if approx_len > counter:
        out += f", ...}} ~{approx_len} entries>"
    else:
        out += "}>"
    return out


def bytesize_repr(num):
    # adapted from dask.utils
    for x in ["B", "KB", "MB", "GB", "TB"]:
        if num < 1024.0:
            if x == "B":
                s = f"{num:.0f} {x}"
            else:
                s = f"{num:.1f} {x}"
            return s
        num /= 1024.0


def is_coroutine_callable(call: Callable[..., Any]) -> bool:
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    dunder_call = getattr(call, "__call__", None)  # noqa: B004
    return inspect.iscoroutinefunction(dunder_call)


async def ensure_awaitable(func, *args, **kwargs):
    if is_coroutine_callable(func):
        return await func(*args, **kwargs)
    else:
        return await anyio.to_thread.run_sync(func, *args, **kwargs)


def path_from_uri(uri):
    """
    Give a URI, return a Path.

    If the URI has a scheme other than 'file', raise ValueError.

    >>> path_from_uri('file://localhost/a/b/c')  # POSIX-style
    '/a/b/c'
    >>> path_from_uri('file://localhost/C:/a/b/c')  # Windows-style
    'C:/a/b/c'
    """
    parsed = urlparse(uri)
    if parsed.scheme != "file":
        raise ValueError(f"Only 'file' URIs are supported. URI: {uri}")
    if platform.system() == "Windows":
        # We slice because we need "C:/..." not "/C:/...".
        path = Path(parsed.path[1:])
    else:
        path = Path(parsed.path)
    return path



import pytest

from ..adapters.mapping import MapAdapter
from ..client import Context, from_context, record_history
from ..server.app import build_app


@pytest.fixture(scope="module")
def context():
    tree = MapAdapter({})
    app = build_app(tree)
    with Context.from_app(app) as context:
        yield context


def test_history(context):
    "Very basic exercise of history"
    client = from_context(context)
    with record_history() as history:
        repr(client)  # trigger a request
    assert history.requests
    assert history.responses



import io
import os
import shutil
import subprocess
import sys
import time

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..client.auth import CannotRefreshAuthentication
from ..client.context import clear_default_identity, get_default_identity
from ..server import authentication
from ..server.app import build_app_from_config
from .utils import fail_with_status_code

arr = ArrayAdapter.from_array(numpy.ones((5, 5)))


tree = MapAdapter({"A1": arr, "A2": arr})


@pytest.fixture
def config(tmpdir):
    """
    Return config with

    - a unique temporary sqlite database location
    - a unique nested dict instance that the test can mutate
    """
    database_uri = f"sqlite+aiosqlite:///{tmpdir}/tiled.sqlite"
    subprocess.run(
        [sys.executable, "-m", "tiled", "admin", "initialize-database", database_uri],
        check=True,
        capture_output=True,
    )
    return {
        "authentication": {
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {"alice": "secret1", "bob": "secret2"}
                    },
                }
            ],
        },
        "database": {
            "uri": database_uri,
        },
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/",
            },
        ],
    }


def test_password_auth(enter_password, config):
    """
    A password that is wrong, empty, or belonging to a different user fails.
    """
    with Context.from_app(build_app_from_config(config)) as context:
        # Log in as Alice.
        with enter_password("secret1"):
            from_context(context, username="alice")
        # Reuse token from cache.
        client = from_context(context, username="alice")
        client.logout()

        # Log in as Bob.
        with enter_password("secret2"):
            client = from_context(context, username="bob")
        client.logout()

        # Bob's password should not work for Alice.
        with fail_with_status_code(401):
            with enter_password("secret2"):
                from_context(context, username="alice")

        # Empty password should not work.
        with fail_with_status_code(422):
            with enter_password(""):
                from_context(context, username="alice")


def test_logout(enter_password, config, tmpdir):
    """
    Logging out revokes the session, such that it cannot be refreshed.
    """
    with Context.from_app(build_app_from_config(config)) as context:
        # Log in as Alice.
        with enter_password("secret1"):
            from_context(context, username="alice")
        # Reuse token from cache.
        client = from_context(context, username="alice")
        # This was set to a unique (temporary) dir by an autouse fixture in conftest.py.
        tiled_cache_dir = os.environ["TILED_CACHE_DIR"]
        # Make a backup copy of the cache directory, which contains the auth tokens.
        shutil.copytree(tiled_cache_dir, tmpdir / "backup")
        # Logout does two things:
        # 1. Revoke the session, so that it cannot be refreshed.
        # 2. Clear the tokens related to this session from in-memory state
        #    and on-disk state.
        client.logout()
        # Restore the tokens from backup.
        # Our aim is to test, below, that even if you have the tokens they
        # can't be used anymore.
        shutil.rmtree(tiled_cache_dir)
        shutil.copytree(tmpdir / "backup", tiled_cache_dir)
        # There is no way to revoke a JWT access token. It expires after a
        # short time window (minutes) but it will still work here, as it has
        # not been that long.
        client = from_context(context, username="alice")
        # The refresh token refers to a revoked session, so refreshing the
        # session to generate a *new* access and refresh token will fail.
        with pytest.raises(CannotRefreshAuthentication):
            client.context.force_auth_refresh()


def test_key_rotation(enter_password, config):
    """
    Rotate in a new secret used to sign keys.
    Confirm that clients experience a smooth transition.
    """

    with Context.from_app(build_app_from_config(config)) as context:
        # Obtain refresh token.
        with enter_password("secret1"):
            from_context(context, username="alice")
        # Use refresh token (no prompt to reauthenticate).
        client = from_context(context, username="alice")

    # Rotate in a new key.
    config["authentication"]["secret_keys"].insert(0, "NEW_SECRET")
    assert config["authentication"]["secret_keys"] == ["NEW_SECRET", "SECRET"]

    with Context.from_app(build_app_from_config(config)) as context:
        # The refresh token from the old key is still valid. No login prompt here.
        client = from_context(context, username="alice")
        # We reauthenticate and receive a refresh token for the new key.
        # (This would happen on its own with the passage of time, but we force it
        # for the sake of a quick test.)
        client.context.force_auth_refresh()

    # Rotate out the old key.
    del config["authentication"]["secret_keys"][1]
    assert config["authentication"]["secret_keys"] == ["NEW_SECRET"]

    with Context.from_app(build_app_from_config(config)) as context:
        # New refresh token works with the new key
        from_context(context, username="alice")


def test_refresh_forced(enter_password, config):
    "Forcing refresh obtains new token."
    with Context.from_app(build_app_from_config(config)) as context:
        # Normal default configuration: a refresh is not immediately required.
        with enter_password("secret1"):
            client = from_context(context, username="alice")
        tokens1 = dict(client.context.tokens)
        # Wait for a moment or we will get a new token that happens to be identical
        # to the old token. This advances the expiration time to make a distinct token.
        time.sleep(2)
        # Forcing a refresh gives us a new token.
        client.context.force_auth_refresh()
        tokens2 = dict(client.context.tokens)
        assert tokens1 != tokens2


def test_refresh_transparent(enter_password, config):
    "When access token expired, refresh happens transparently."
    # Pathological configuration: a refresh is almost immediately required
    config["authentication"]["access_token_max_age"] = 1
    with Context.from_app(build_app_from_config(config)) as context:
        with enter_password("secret1"):
            client = from_context(context, username="alice")
        tokens1 = dict(client.context.tokens)
        time.sleep(2)
        # A refresh should happen automatically now.
        client["A1"]
        tokens2 = dict(client.context.tokens)
        assert tokens2 != tokens1


def test_expired_session(enter_password, config):
    # Pathological configuration: sessions do not last
    config["authentication"]["session_max_age"] = 1
    with Context.from_app(build_app_from_config(config)) as context:
        with enter_password("secret1"):
            client = from_context(context, username="alice")
        time.sleep(2)
        # Refresh should fail because the session is too old.
        with pytest.raises(CannotRefreshAuthentication):
            client.context.force_auth_refresh()


def test_revoke_session(enter_password, config):
    with Context.from_app(build_app_from_config(config)) as context:
        with enter_password("secret1"):
            client = from_context(context, username="alice")
        # Get the current session ID.
        info = client.context.whoami()
        (session,) = info["sessions"]
        assert not session["revoked"]
        # Revoke it.
        client.context.revoke_session(session["uuid"])
        # Update info and confirm it is listed as revoked.
        updated_info = client.context.whoami()
        (updated_session,) = updated_info["sessions"]
        assert updated_session["revoked"]
        # Confirm it cannot be refreshed.
        with pytest.raises(CannotRefreshAuthentication):
            client.context.force_auth_refresh()


def test_multiple_providers(enter_password, config, monkeypatch):
    """
    Test a configuration with multiple identity providers.

    This mechanism is used to support "Login with ORCID or Google or ...."
    """
    config["authentication"]["providers"].extend(
        [
            {
                "provider": "second",
                "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                "args": {"users_to_passwords": {"cara": "secret3", "doug": "secret4"}},
            },
            {
                "provider": "third",
                "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                "args": {
                    # Duplicate 'cara' username.
                    "users_to_passwords": {"cara": "secret5", "emilia": "secret6"}
                },
            },
        ],
    )
    with Context.from_app(build_app_from_config(config)) as context:
        monkeypatch.setattr("sys.stdin", io.StringIO("1\n"))
        with enter_password("secret1"):
            from_context(context, username="alice")
        monkeypatch.setattr("sys.stdin", io.StringIO("2\n"))
        with enter_password("secret3"):
            from_context(context, username="cara")
        monkeypatch.setattr("sys.stdin", io.StringIO("3\n"))
        with enter_password("secret5"):
            from_context(context, username="cara")


def test_multiple_providers_name_collision(config):
    """
    Check that we enforce unique provider names.
    """
    config["authentication"]["providers"] = [
        {
            "provider": "some_name",
            "authenticator": "tiled.authenticators:DictionaryAuthenticator",
            "args": {"users_to_passwords": {"cara": "secret3", "doug": "secret4"}},
        },
        {
            "provider": "some_name",  # duplicate!
            "authenticator": "tiled.authenticators:DictionaryAuthenticator",
            "args": {
                # Duplicate 'cara' username.
                "users_to_passwords": {"cara": "secret5", "emilia": "secret6"}
            },
        },
    ]
    with pytest.raises(ValueError):
        build_app_from_config(config)


def test_admin(enter_password, config):
    """
    Test that the 'tiled_admin' config confers the 'admin' Role on a Principal.
    """
    # Make alice an admin. Leave bob as a user.
    config["authentication"]["tiled_admins"] = [{"provider": "toy", "id": "alice"}]

    with Context.from_app(build_app_from_config(config)) as context:
        with enter_password("secret1"):
            context.authenticate(username="alice")
        admin_roles = context.whoami()["roles"]
        assert "admin" in [role["name"] for role in admin_roles]

        # Exercise admin functions.
        principals = context.admin.list_principals()
        some_principal_uuid = principals[0]["uuid"]
        context.admin.show_principal(some_principal_uuid)

        with enter_password("secret2"):
            context.authenticate(username="bob")
        user_roles = context.whoami()["roles"]
        assert [role["name"] for role in user_roles] == ["user"]

        # As bob, admin functions should be disallowed.
        with fail_with_status_code(401):
            context.admin.list_principals()
        with fail_with_status_code(401):
            context.admin.show_principal(some_principal_uuid)


def test_api_key_activity(enter_password, config):
    """
    Create and use an API. Verify that latest_activity updates.
    """
    with Context.from_app(build_app_from_config(config)) as context:
        # Log in as user.
        with enter_password("secret1"):
            context.authenticate(username="alice")
        # Make and use an API key. Check that latest_activity is not set.
        key_info = context.create_api_key()
        context.logout()
        assert key_info["latest_activity"] is None  # never used
        context.api_key = key_info["secret"]

        # Use the key for a couple requests and see that latest_activity becomes set and then increases.
        client = from_context(context)
        client["A1"]
        key_activity1 = context.which_api_key()["latest_activity"]
        principal_activity1 = context.whoami()["latest_activity"]
        assert key_activity1 is not None
        time.sleep(2)  # Ensure time resolution (1 second) has ticked up.
        client["A1"]
        key_activity2 = context.which_api_key()["latest_activity"]
        principal_activity2 = context.whoami()["latest_activity"]
        assert key_activity2 > key_activity1
        assert principal_activity2 > principal_activity1
        assert len(context.whoami()["api_keys"]) == 1

        # Unset the API key.
        secret = context.api_key
        context.api_key = None
        with pytest.raises(RuntimeError):
            context.which_api_key()
        # Set the API key.
        context.api_key = secret
        # Now this works again.
        context.which_api_key()


def test_api_key_scopes(enter_password, config):
    # Make alice an admin. Leave bob as a user.
    config["authentication"]["tiled_admins"] = [{"provider": "toy", "id": "alice"}]
    with Context.from_app(build_app_from_config(config)) as context:
        # Log in as admin.
        with enter_password("secret1"):
            context.authenticate(username="alice")
        # Request a key with reduced scope that cannot read metadata.
        metrics_key_info = context.create_api_key(scopes=["metrics"])
        context.logout()
        context.api_key = metrics_key_info["secret"]
        with fail_with_status_code(401):
            from_context(context)
        context.api_key = None

        # Log in as ordinary user.
        with enter_password("secret2"):
            context.authenticate(username="bob")
        # Try to request a key with more scopes that the user has.
        with fail_with_status_code(400):
            context.create_api_key(scopes=["admin:apikeys"])
        # Request a key with reduced scope that can *only* read metadata.
        metadata_key_info = context.create_api_key(scopes=["read:metadata"])
        context.logout()
        context.api_key = metadata_key_info["secret"]
        restricted_client = from_context(context)
        restricted_client["A1"]
        with fail_with_status_code(401):
            restricted_client["A1"].read()  # no 'read:data' scope
        context.api_key = None


def test_api_key_revoked(enter_password, config):
    with Context.from_app(build_app_from_config(config)) as context:
        with enter_password("secret1"):
            context.authenticate(username="alice")

        # Create a key with a note.
        NOTE = "will revoke soon"
        key_info = context.create_api_key(note=NOTE)
        assert len(context.whoami()["api_keys"]) == 1
        (api_key,) = context.whoami()["api_keys"]
        assert api_key["note"] == key_info["note"] == NOTE
        assert api_key["first_eight"] == key_info["first_eight"]
        context.logout()

        # Use it.
        context.api_key = key_info["secret"]
        from_context(context)
        context.api_key = None

        # Revoke the new key.
        with enter_password("secret1"):
            context.authenticate(username="alice")
        context.revoke_api_key(key_info["first_eight"])
        assert len(context.whoami()["api_keys"]) == 0
        context.logout()
        context.api_key = key_info["secret"]
        with fail_with_status_code(401):
            from_context(context)


def test_api_key_expiration(enter_password, config):
    with Context.from_app(build_app_from_config(config)) as context:
        with enter_password("secret1"):
            context.authenticate(username="alice")
        # Create a key with a very short lifetime.
        key_info = context.create_api_key(
            note="will expire very soon", expires_in=1
        )  # units: seconds
        context.logout()
        context.api_key = key_info["secret"]
        time.sleep(2)
        with fail_with_status_code(401):
            from_context(context)


def test_api_key_limit(enter_password, config):
    # Decrease the limit so this test runs faster.
    original_limit = authentication.API_KEY_LIMIT
    authentication.API_KEY_LIMIT = 3
    try:
        with Context.from_app(build_app_from_config(config)) as context:
            with enter_password("secret2"):
                context.authenticate(username="bob")
            for i in range(authentication.API_KEY_LIMIT):
                context.create_api_key(note=f"key {i}")
            # Hit API key limit.
            with fail_with_status_code(400):
                context.create_api_key(note="one key too many")
    finally:
        authentication.API_KEY_LIMIT = original_limit


def test_session_limit(enter_password, config):
    # Decrease the limit so this test runs faster.
    original_limit = authentication.SESSION_LIMIT
    authentication.SESSION_LIMIT = 3
    # Use separate token caches to de-couple login attempts into separate sessions.
    try:
        with Context.from_app(build_app_from_config(config)) as context:
            with enter_password("secret1"):
                for i in range(authentication.SESSION_LIMIT):
                    context.authenticate(username="alice")
                    context.logout()
                # Hit Session limit.
                with fail_with_status_code(400):
                    context.authenticate(username="alice")
    finally:
        authentication.SESSION_LIMIT = original_limit


def test_sticky_identity(enter_password, config):
    # Log in as Alice.
    with Context.from_app(build_app_from_config(config)) as context:
        assert get_default_identity(context.api_uri) is None
        with enter_password("secret1"):
            context.authenticate(username="alice")
        assert context.whoami()["identities"][0]["id"] == "alice"
    # The default identity is now set. The login was "sticky".
    with Context.from_app(build_app_from_config(config)) as context:
        assert get_default_identity(context.api_uri) is not None
        context.authenticate()
        assert context.whoami()["identities"][0]["id"] == "alice"
    # Opt out of the stickiness (set_default=False).
    with Context.from_app(build_app_from_config(config)) as context:
        assert get_default_identity(context.api_uri) is not None
        with enter_password("secret2"):
            context.authenticate(username="bob", set_default=False)
        assert context.whoami()["identities"][0]["id"] == "bob"
    # The default is still Alice.
    with Context.from_app(build_app_from_config(config)) as context:
        assert get_default_identity(context.api_uri) is not None
        context.authenticate()
        assert context.whoami()["identities"][0]["id"] == "alice"
    # Clear the default.
    clear_default_identity(context.api_uri)
    assert get_default_identity(context.api_uri) is None


@pytest.fixture
def principals_context(enter_password, config):
    """
    Fetch UUID for an admin and an ordinary user; include the client context.
    """
    # Make alice an admin. Leave bob as a user.
    config["authentication"]["tiled_admins"] = [{"provider": "toy", "id": "alice"}]

    with Context.from_app(build_app_from_config(config)) as context:
        # Log in as Alice and retrieve admin UUID for later use
        with enter_password("secret1"):
            context.authenticate(username="alice")

        principal = context.whoami()
        assert "admin" in (role["name"] for role in principal["roles"])
        admin_uuid = principal["uuid"]
        context.logout()

        # Log in as Bob and retrieve Bob's UUID for later use
        with enter_password("secret2"):
            context.authenticate(username="bob")

        principal = context.whoami()
        assert "admin" not in (role["name"] for role in principal["roles"])
        bob_uuid = principal["uuid"]
        context.logout()

        yield {
            "uuid": {"alice": admin_uuid, "bob": bob_uuid},
            "context": context,
        }


@pytest.mark.parametrize(
    "username, scopes, resource",
    (
        ("alice", ["read:principals"], "/api/v1/auth/principal"),
        ("bob", ["read:data"], "/api/v1/array/full/A1"),
    ),
)
def test_admin_api_key_any_principal(
    enter_password, principals_context, username, scopes, resource
):
    """
    Admin can create usable API keys for any prinicipal, within that principal's scopes.
    """
    with principals_context["context"] as context:
        # Log in as Alice, create and use API key after logout
        with enter_password("secret1"):
            context.authenticate(username="alice")

        principal_uuid = principals_context["uuid"][username]
        api_key_info = context.admin.create_api_key(principal_uuid, scopes=scopes)
        api_key = api_key_info["secret"]
        assert api_key
        context.logout()

        context.api_key = api_key
        context.http_client.get(resource).raise_for_status()
        context.api_key = None
        # The same endpoint fails without an API key
        with fail_with_status_code(401):
            context.http_client.get(resource).raise_for_status()


def test_admin_create_service_principal(enter_password, principals_context):
    """
    Admin can create service accounts with API keys.
    """
    with principals_context["context"] as context:
        # Log in as Alice, create and use API key after logout
        with enter_password("secret1"):
            context.authenticate(username="alice")

        assert context.whoami()["type"] == "user"

        principal_info = context.admin.create_service_principal(role="user")
        principal_uuid = principal_info["uuid"]

        service_api_key_info = context.admin.create_api_key(principal_uuid)
        context.logout()

        context.api_key = service_api_key_info["secret"]
        assert context.whoami()["type"] == "service"


def test_admin_api_key_any_principal_exceeds_scopes(enter_password, principals_context):
    """
    Admin cannot create API key that exceeds scopes for another principal.
    """
    with principals_context["context"] as context:
        # Log in as Alice, create and use API key after logout
        with enter_password("secret1"):
            context.authenticate(username="alice")

        principal_uuid = principals_context["uuid"]["bob"]
        with fail_with_status_code(400) as fail_info:
            context.admin.create_api_key(principal_uuid, scopes=["read:principals"])
        fail_message = " must be a subset of the principal's scopes "
        assert fail_message in fail_info.value.response.text
        context.logout()


@pytest.mark.parametrize("username", ("alice", "bob"))
def test_api_key_any_principal(enter_password, principals_context, username):
    """
    Ordinary user cannot create API key for another principal.
    """
    with principals_context["context"] as context:
        # Log in as Bob, this API endpoint is unauthorized
        with enter_password("secret2"):
            context.authenticate(username="bob")

        principal_uuid = principals_context["uuid"][username]
        with fail_with_status_code(401):
            context.admin.create_api_key(principal_uuid, scopes=["read:metadata"])


def test_api_key_bypass_scopes(enter_password, principals_context):
    """
    Ordinary user cannot create API key that bypasses a scopes restriction.
    """
    with principals_context["context"] as context:
        # Log in as Bob, create API key with empty scopes
        with enter_password("secret2"):
            context.authenticate(username="bob")

        response = context.http_client.post(
            "/api/v1/auth/apikey", json={"expires_in": None, "scopes": []}
        )
        response.raise_for_status()
        api_key = response.json()["secret"]
        assert api_key
        context.logout()

        # Try the new API key with admin and normal resources
        for resource in ("/api/v1/auth/principal", "/api/v1/array/full/A1"):
            # Try with/without key, with/without empty scopes
            for query_params in (
                {"api_key": api_key},
                {"scopes": []},
                {"api_key": api_key, "scopes": []},
            ):
                context.api_key = query_params.pop("api_key", None)
                with fail_with_status_code(401):
                    context.http_client.get(
                        resource, params=query_params
                    ).raise_for_status()



from pathlib import Path

import dask.dataframe

from ..structures.core import StructureFamily
from ..utils import path_from_uri
from .dataframe import DataFrameAdapter


class ParquetDatasetAdapter:
    structure_family = StructureFamily.table

    def __init__(
        self,
        data_uris,
        structure,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        # TODO Store data_uris instead and generalize to non-file schemes.
        self._partition_paths = [path_from_uri(uri) for uri in data_uris]
        self._metadata = metadata or {}
        self._structure = structure
        self.specs = list(specs or [])
        self.access_policy = access_policy

    def metadata(self):
        return self._metadata

    @property
    def dataframe_adapter(self):
        partitions = []
        for path in self._partition_paths:
            if not Path(path).exists():
                partition = None
            else:
                partition = dask.dataframe.read_parquet(path)
            partitions.append(partition)
        return DataFrameAdapter(partitions, self._structure)

    @classmethod
    def init_storage(cls, data_uri, structure):
        from ..server.schemas import Asset

        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        assets = [
            Asset(
                data_uri=f"{data_uri}/partition-{i}.parquet",
                is_directory=False,
                parameter="data_uris",
                num=i,
            )
            for i in range(structure.npartitions)
        ]
        return assets

    def write_partition(self, data, partition):
        uri = self._partition_paths[partition]
        data.to_parquet(uri)

    def write(self, data):
        if self.structure().npartitions != 1:
            raise NotImplementedError
        uri = self._partition_paths[0]
        data.to_parquet(uri)

    def read(self, *args, **kwargs):
        return self.dataframe_adapter.read(*args, **kwargs)

    def read_partition(self, *args, **kwargs):
        return self.dataframe_adapter.read_partition(*args, **kwargs)

    def structure(self):
        return self._structure



import warnings

# for back-compat
from ..utils import node_repr as tree_repr  # noqa: F401

_MESSAGE = (
    "Instead of {name}_indexer[...] use {name}()[...]. "
    "The {name}_indexer accessor is deprecated."
)


class IndexersMixin:
    """
    Provides slicable attributes keys_indexer, items_indexer, values_indexer.

    This is just for back-ward compatiblity.
    """

    @property
    def keys_indexer(self):
        warnings.warn(_MESSAGE.format(name="keys"), DeprecationWarning)
        return self.keys()

    @property
    def values_indexer(self):
        warnings.warn(_MESSAGE.format(name="values"), DeprecationWarning)
        return self.values()

    @property
    def items_indexer(self):
        warnings.warn(_MESSAGE.format(name="items"), DeprecationWarning)
        return self.items()


class IndexCallable:
    """
    DEPRECATED and no longer used internally

    Provide getitem syntax for functions

    >>> def inc(x):
    ...     return x + 1

    >>> I = IndexCallable(inc)
    >>> I[3]
    4

    Vendored from dask
    """

    __slots__ = ("fn",)

    def __init__(self, fn):
        self.fn = fn

    def __getitem__(self, key):
        return self.fn(key)



"""
This module was created to save structure methods that can be handled by
the server and the client.

"""

import enum
from dataclasses import asdict, dataclass
from typing import Optional


class StructureFamily(str, enum.Enum):
    awkward = "awkward"
    container = "container"
    array = "array"
    sparse = "sparse"
    table = "table"


@dataclass(frozen=True)
class Spec:
    name: str
    version: Optional[str] = None

    def __init__(self, name, version=None):
        # Enable the name to be passed as a position argument.
        # The setattr stuff is necessary to make this work with a frozen dataclass.
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "version", version)

    def __repr__(self):
        # Display the name as a positional argument, for conciseness.
        if self.version is None:
            output = f"{type(self).__name__}({self.name!r})"
        else:
            output = f"{type(self).__name__}({self.name!r}, version={self.version!r})"
        return output

    def dict(self):
        return asdict(self)



import typer

profile_app = typer.Typer()


@profile_app.command("paths")
def profile_paths():
    "List the locations that the client will search for profiles (client-side configuration)."
    from ..profiles import paths

    print("\n".join(str(p) for p in paths))


@profile_app.command("list")
def profile_list():
    "List the profiles (client-side configuration) found and the files they were read from."
    from ..profiles import load_profiles

    profiles = load_profiles()
    if not profiles:
        typer.echo("No profiles found.")
        return
    max_len = max(len(name) for name in profiles)
    PADDING = 4

    print(
        "\n".join(
            f"{name:<{max_len + PADDING}}{filepath}"
            for name, (filepath, _) in profiles.items()
        )
    )


@profile_app.command("show")
def profile_show(profile_name: str):
    "Show the content of a profile."
    import sys

    import yaml

    from ..profiles import load_profiles

    profiles = load_profiles()
    try:
        filepath, content = profiles[profile_name]
    except KeyError:
        typer.echo(
            f"The profile {profile_name!r} could not be found. "
            "Use tiled profile list to see profile names.",
            err=True,
        )
        raise typer.Abort()
    print(f"Source: {filepath}", file=sys.stderr)
    print("--", file=sys.stderr)
    print(yaml.dump(content), file=sys.stdout)


@profile_app.command("edit")
def profile_edit(profile_name: str):
    "Show the content of a profile."
    import sys

    from ..profiles import load_profiles

    profiles = load_profiles()
    try:
        filepath, content = profiles[profile_name]
    except KeyError:
        typer.echo(
            f"The profile {profile_name!r} could not be found. "
            "Use tiled profile list to see profile names.",
            err=True,
        )
        raise typer.Abort()
    print(f"Opening {filepath} in default text editor...", file=sys.stderr)

    import os
    import platform
    import subprocess

    if platform.system() == "Darwin":
        subprocess.call(("open", filepath))
    elif platform.system() == "Windows":
        os.startfile(filepath)
    else:
        subprocess.call(("xdg-open", filepath))


@profile_app.command("create")
def create(
    uri: str = typer.Argument(..., help="URI 'http[s]://...'"),
    name: str = typer.Option("auto", help="Profile name, a short convenient alias"),
    set_default: bool = typer.Option(
        True, help="Set new profile as the default profile."
    ),
    overwrite: bool = typer.Option(
        False, "--overwrite", help="Overwrite an existing profile of this name."
    ),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    """
    Create a 'profile' that can be used to connect to a Tiled server.
    """
    from ..profiles import ProfileExists, create_profile, set_default_profile_name

    try:
        create_profile(name=name, uri=uri, verify=not no_verify, overwrite=overwrite)
    except ProfileExists:
        typer.echo(
            f"A profile named {name!r} already exists. Use --overwrite to overwrite it."
        )
        raise typer.Abort()
    if set_default:
        set_default_profile_name(name)
        typer.echo(f"Tiled profile {name!r} created and set as the default.", err=True)
    else:
        typer.echo(f"Tiled profile {name!r} created.", err=True)


@profile_app.command("delete")
def delete(
    name: str = typer.Argument(..., help="Profile name"),
):
    from ..profiles import (
        delete_profile,
        get_default_profile_name,
        set_default_profile_name,
    )

    # Unset the default if this profile is currently the default.
    default = get_default_profile_name()
    if default == name:
        set_default_profile_name(None)
    delete_profile(name)
    typer.echo(f"Tiled profile {name!r} deleted.", err=True)


@profile_app.command("get-default")
def get_default():
    """
    Show the current default Tiled profile.
    """
    from ..profiles import get_default_profile_name, load_profiles

    name = get_default_profile_name()
    if name is None:
        typer.echo("No default.", err=True)
    else:
        import yaml

        source_filepath, profile_content = load_profiles()[name]
        typer.echo(f"# Profile name: {name!r}")
        typer.echo(f"# {source_filepath} \n")
        typer.echo(yaml.dump(profile_content))


@profile_app.command("set-default")
def set_default(profile_name: str):
    """
    Set the default Tiled profile.
    """
    from ..profiles import set_default_profile_name

    set_default_profile_name(profile_name)


@profile_app.command("clear-default")
def clear_default():
    """
    Clear the default Tiled profile.
    """
    from ..profiles import set_default_profile_name

    set_default_profile_name(None)



import collections.abc
import dataclasses
import itertools
import math
import operator
import re
import sys
import types
import uuid
from collections import defaultdict
from datetime import datetime, timedelta
from hashlib import md5
from typing import Any

import anyio
import dateutil.tz
import jmespath
import msgpack
from fastapi import HTTPException, Response
from starlette.responses import JSONResponse, Send, StreamingResponse

# Some are not directly used, but they register things on import.
from .. import queries
from ..adapters.mapping import MapAdapter
from ..queries import KeyLookup, QueryValueError
from ..serialization import register_builtin_serializers
from ..structures.core import Spec, StructureFamily
from ..utils import (
    APACHE_ARROW_FILE_MIME_TYPE,
    SerializationError,
    UnsupportedShape,
    ensure_awaitable,
    safe_json_dump,
)
from . import schemas
from .etag import tokenize
from .utils import record_timing

del queries
register_builtin_serializers()


_FILTER_PARAM_PATTERN = re.compile(r"filter___(?P<name>.*)___(?P<field>[^\d\W][\w\d]+)")
_LOCAL_TZINFO = dateutil.tz.gettz()

# Pragmatic limit on how "wide" a node can be before the server refuses to
# inline its contents.
# There are some wide-and-short DataFrame datasets we have observed
# (specifically, bluesky "baseline" readings) that can touch this limit.
# We raised it from an initial guess of 100 to avoid splitting these over
# too many requests. This value *may* be too high --- need more benchmarking
# on realistic workloads.
INLINED_CONTENTS_LIMIT = 500

# Pragmatic limit on how deep the server will recurse into nodes that request
# inlined contents. This is a hard upper bound meant to protect the server from
# being crashed by badly designed or buggy Adapters. It is up to Adapters to
# opt in to this behavior and decide on a reasonable depth.
DEPTH_LIMIT = 5

DEFAULT_PAGE_SIZE = 100
MAX_PAGE_SIZE = 300


async def len_or_approx(tree):
    """
    Prefer approximate length if implemented. (It's cheaper.)
    """
    if hasattr(tree, "async_len"):
        return await tree.async_len()
    try:
        return await anyio.to_thread.run_sync(operator.length_hint, tree)
    except TypeError:
        return await anyio.to_thread.run_sync(len, tree)


def pagination_links(base_url, route, path_parts, offset, limit, length_hint):
    path_str = "/".join(path_parts)
    links = {
        "self": f"{base_url}{route}/{path_str}?page[offset]={offset}&page[limit]={limit}",
        # These are conditionally overwritten below.
        "first": None,
        "last": None,
        "next": None,
        "prev": None,
    }
    if limit:
        last_page = math.floor(length_hint / limit) * limit
        links.update(
            {
                "first": f"{base_url}{route}/{path_str}?page[offset]={0}&page[limit]={limit}",
                "last": f"{base_url}{route}/{path_str}?page[offset]={last_page}&page[limit]={limit}",
            }
        )
    if offset + limit < length_hint:
        links[
            "next"
        ] = f"{base_url}{route}/{path_str}?page[offset]={offset + limit}&page[limit]={limit}"
    if offset > 0:
        links[
            "prev"
        ] = f"{base_url}{route}/{path_str}?page[offset]={max(0, offset - limit)}&page[limit]={limit}"
    return links


async def apply_search(tree, filters, query_registry):
    queries = defaultdict(
        dict
    )  # e.g. {"text": {"text": "dog"}, "lookup": {"key": "..."}}
    # Group the parameters by query type.
    for key, value in filters.items():
        if value is None:
            continue
        name, field = _FILTER_PARAM_PATTERN.match(key).groups()
        queries[name][field] = value

    # Apply the queries and obtain a narrowed tree.
    key_lookups = []
    for query_name, parameters_dict_of_lists in queries.items():
        for i in itertools.count(0):
            try:
                parameters = {
                    field_name: parameters_list[i]
                    for field_name, parameters_list in parameters_dict_of_lists.items()
                }
            except IndexError:
                break
            query_class = query_registry.name_to_query_type[query_name]
            try:
                query = query_class.decode(**parameters)
                # Special case: Do key-lookups at the end after all other filtering.
                # We do not require trees to implement this query; we implement it
                # directly here by just calling __getitem__.
                if isinstance(query, KeyLookup):
                    key_lookups.append(query.key)
                    continue
                tree = tree.search(query)
            except QueryValueError as err:
                raise HTTPException(status_code=400, detail=err.args[0])
    if key_lookups:
        # Duplicates are technically legal because *any* query can be given
        # with multiple parameters.
        unique_key_lookups = set(key_lookups)
        (key_lookup), *others = unique_key_lookups
        if others:
            # Two non-equal KeyLookup queries must return no results.
            tree = MapAdapter({})
        else:
            if hasattr(tree, "lookup_adapter"):
                entry = await tree.lookup_adapter([key_lookup])
                if entry is None:
                    tree = MapAdapter({})
                else:
                    tree = MapAdapter({key_lookup: entry}, must_revalidate=False)
            else:
                try:
                    tree = MapAdapter(
                        {key_lookup: tree[key_lookup]}, must_revalidate=False
                    )
                except KeyError:
                    tree = MapAdapter({})

    return tree


def apply_sort(tree, sort):
    sorting = []
    if sort is not None:
        for item in sort.split(","):
            if item:
                if item.startswith("-"):
                    sorting.append((item[1:], -1))
                else:
                    sorting.append((item, 1))
    if sorting:
        if not hasattr(tree, "sort"):
            raise HTTPException(
                status_code=400, detail="This Tree does not support sorting."
            )
        tree = tree.sort(sorting)

    return tree


async def construct_entries_response(
    query_registry,
    tree,
    route,
    path,
    offset,
    limit,
    fields,
    select_metadata,
    omit_links,
    show_sources,
    filters,
    sort,
    base_url,
    media_type,
    max_depth,
):
    path_parts = [segment for segment in path.split("/") if segment]
    tree = await apply_search(tree, filters, query_registry)
    tree = apply_sort(tree, sort)

    count = await len_or_approx(tree)
    links = pagination_links(base_url, route, path_parts, offset, limit, count)
    data = []
    if fields != [schemas.EntryFields.none]:
        # Pull a page of items into memory.
        if hasattr(tree, "items_range"):
            items = await tree.items_range(offset, limit)
        else:
            items = tree.items()[offset : offset + limit]  # noqa: E203
    else:
        # Pull a page of just the keys, which is cheaper.
        if hasattr(tree, "keys_range"):
            keys = await tree.keys_range(offset, limit)
        else:
            keys = tree.keys()[offset : offset + limit]  # noqa: E203
        items = [(key, None) for key in keys]
    # This value will not leak out. It just used to seed comparisons.
    metadata_stale_at = datetime.utcnow() + timedelta(days=1_000_000)
    must_revalidate = getattr(tree, "must_revalidate", True)
    for key, entry in items:
        resource = await construct_resource(
            base_url,
            path_parts + [key],
            entry,
            fields,
            select_metadata,
            omit_links,
            show_sources,
            media_type,
            max_depth=max_depth,
        )
        data.append(resource)
        # If any entry has emtry.metadata_stale_at = None, then there will
        # be no 'Expires' header. We will pessimistically assume the values
        # are immediately stale.
        if metadata_stale_at is not None:
            if getattr(entry, "metadata_stale_at", None) is None:
                metadata_stale_at = None
            else:
                metadata_stale_at = min(metadata_stale_at, entry.metadata_stale_at)
    return (
        schemas.Response(data=data, links=links, meta={"count": count}),
        metadata_stale_at,
        must_revalidate,
    )


DEFAULT_MEDIA_TYPES = {
    StructureFamily.array: {"*/*": "application/octet-stream", "image/*": "image/png"},
    StructureFamily.awkward: {"*/*": "application/zip"},
    StructureFamily.table: {"*/*": APACHE_ARROW_FILE_MIME_TYPE},
    StructureFamily.container: {"*/*": "application/x-hdf5"},
    StructureFamily.sparse: {"*/*": APACHE_ARROW_FILE_MIME_TYPE},
}


async def construct_revisions_response(
    entry,
    base_url,
    route,
    path,
    offset,
    limit,
    media_type,
):
    path_parts = [segment for segment in path.split("/") if segment]
    revisions = await entry.revisions(offset, limit)
    data = []
    for revision in revisions:
        item = {
            "revision_number": revision.revision_number,
            "attributes": {
                "metadata": revision.metadata,
                "specs": revision.specs,
                "time_updated": revision.time_updated,
            },
        }
        data.append(item)
    count = len(data)
    links = pagination_links(
        base_url, route, path_parts, offset, limit, count
    )  # maybe reuse or maybe make a new pagination_revision_links
    return schemas.Response(data=data, links=links, meta={"count": count})


async def construct_data_response(
    structure_family,
    serialization_registry,
    payload,
    metadata,
    request,
    format=None,
    specs=None,
    expires=None,
    filename=None,
    filter_for_access=None,
):
    request.state.endpoint = "data"
    if specs is None:
        specs = []
    default_media_type = DEFAULT_MEDIA_TYPES[structure_family]["*/*"]
    # Give priority to the `format` query parameter. Otherwise, consult Accept
    # header.
    if format is not None:
        media_types_or_aliases = format.split(",")
        # Resolve aliases, like "csv" -> "text/csv".
        media_types = [
            serialization_registry.resolve_alias(t) for t in media_types_or_aliases
        ]
    else:
        # The HTTP spec says these should be separated by ", " but some
        # browsers separate with just "," (no space).
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Content_negotiation/List_of_default_Accept_values#default_values  # noqa
        # That variation is what we are handling below with lstrip.
        media_types = [
            s.lstrip(" ")
            for s in request.headers.get("Accept", default_media_type).split(",")
        ]

    # The client may give us a choice of media types. Find the first one
    # that we support.
    supported = set()
    for media_type in media_types:
        if media_type == "*/*":
            media_type = DEFAULT_MEDIA_TYPES[structure_family]["*/*"]
        elif structure_family == StructureFamily.array and media_type == "image/*":
            media_type = DEFAULT_MEDIA_TYPES[structure_family]["image/*"]
        # Compare the request formats to the formats supported by each spec
        # name and, finally, by the structure family.
        for spec in [spec.name for spec in specs] + [structure_family]:
            media_types_for_spec = serialization_registry.media_types(spec)
            if media_type in media_types_for_spec:
                break
            supported.update(media_types_for_spec)
        else:
            # None of the specs or the structure_family can serialize to this
            # media_type. Try the next one.
            continue
        # We found a match above. We have our media_type.
        break
    else:
        # We have checked each of the media_types, and we cannot serialize
        # to any of them.
        raise UnsupportedMediaTypes(
            f"None of the media types requested by the client are supported. "
            f"Supported: {', '.join(supported)}. Requested: {', '.join(media_types)}.",
        )
    with record_timing(request.state.metrics, "tok"):
        # Create an ETag that uniquely identifies this content and the media
        # type that it will be encoded as.
        etag = tokenize((payload, media_type))
    headers = {"ETag": etag}
    if expires is not None:
        headers["Expires"] = expires.strftime(HTTP_EXPIRES_HEADER_FORMAT)
    if request.headers.get("If-None-Match", "") == etag:
        # If the client already has this content, confirm that.
        return Response(status_code=304, headers=headers)
    if filename:
        headers["Content-Disposition"] = f"attachment;filename={filename}"
    serializer = serialization_registry.dispatch(spec, media_type)
    # This is the expensive step: actually serialize.
    try:
        if filter_for_access is not None:
            content = await ensure_awaitable(
                serializer, payload, metadata, filter_for_access
            )
        else:
            content = await ensure_awaitable(serializer, payload, metadata)
    except UnsupportedShape as err:
        raise UnsupportedMediaTypes(
            f"The shape of this data {err.args[0]} is incompatible with the requested format ({media_type}). "
            f"Slice it or choose a different format.",
        )
    except SerializationError as err:
        raise UnsupportedMediaTypes(
            f"This type is supported in general but there was an error packing this specific data: {err.args[0]}",
        )
    if isinstance(content, types.GeneratorType):
        response_class = PatchedStreamingResponse
    else:
        response_class = PatchedResponse
    return response_class(
        content,
        media_type=media_type,
        headers=headers,
    )


async def construct_resource(
    base_url,
    path_parts,
    entry,
    fields,
    select_metadata,
    omit_links,
    show_sources,
    media_type,
    max_depth,
    depth=0,
):
    path_str = "/".join(path_parts)
    attributes = {"ancestors": path_parts[:-1]}
    if show_sources and hasattr(entry, "data_sources"):
        attributes["data_sources"] = entry.data_sources
    if schemas.EntryFields.metadata in fields:
        if select_metadata is not None:
            attributes["metadata"] = jmespath.compile(select_metadata).search(
                entry.metadata()
            )
        else:
            attributes["metadata"] = entry.metadata()
    if schemas.EntryFields.specs in fields:
        specs = []
        for spec in getattr(entry, "specs", []):
            # back-compat for when a spec was just a string
            if isinstance(spec, str):
                spec = Spec(spec)
            # Convert from dataclass to pydantic.
            # The dataclass implementation of Spec supports dict() method
            # for ease of going between dataclass and pydantic.
            specs.append(schemas.Spec(**spec.dict()))
        attributes["specs"] = specs
    if (entry is not None) and entry.structure_family == StructureFamily.container:
        attributes["structure_family"] = StructureFamily.container
        if schemas.EntryFields.structure in fields:
            if (
                ((max_depth is None) or (depth < max_depth))
                and hasattr(entry, "inlined_contents_enabled")
                and entry.inlined_contents_enabled(depth)
                and depth <= DEPTH_LIMIT
            ):
                # This node wants us to inline its contents.
                # First check that it is not too large.
                est_count = await len_or_approx(entry)
                if est_count > INLINED_CONTENTS_LIMIT:
                    # Too large: do not inline its contents.
                    count = est_count
                    contents = None
                else:
                    contents = {}
                    # The size may change as we are walking the entry.
                    # Keep a *true* count separately from est_count.
                    count = 0
                    for key, adapter in entry.items():
                        count += 1
                        if count > INLINED_CONTENTS_LIMIT:
                            # The est_count was inaccurate or else the entry has grown
                            # new children while we are walking it. Too large!
                            count = await len_or_approx(entry)
                            contents = None
                            break
                        contents[key] = await construct_resource(
                            base_url,
                            path_parts + [key],
                            adapter,
                            fields,
                            select_metadata,
                            omit_links,
                            show_sources,
                            media_type,
                            max_depth,
                            depth=1 + depth,
                        )
            else:
                count = await len_or_approx(entry)
                contents = None
            structure = schemas.NodeStructure(
                count=count,
                contents=contents,
            )
            attributes["structure"] = structure
        if schemas.EntryFields.sorting in fields:
            if hasattr(entry, "sorting"):
                # HUGE HACK
                # In the Python API we encode sorting as (key, direction).
                # This order-based "record" notion does not play well with OpenAPI.
                # In the HTTP API, therefore, we use {"key": key, "direction": direction}.
                if entry.sorting and isinstance(entry.sorting[0], schemas.SortingItem):
                    attributes["sorting"] = entry.sorting
                else:
                    attributes["sorting"] = [
                        {"key": key, "direction": direction}
                        for key, direction in entry.sorting
                    ]
        d = {
            "id": path_parts[-1] if path_parts else "",
            "attributes": schemas.NodeAttributes(**attributes),
        }
        if not omit_links:
            d["links"] = {
                "self": f"{base_url}/metadata/{path_str}",
                "search": f"{base_url}/search/{path_str}",
                "full": f"{base_url}/container/full/{path_str}",
            }

        resource = schemas.Resource[
            schemas.NodeAttributes, schemas.ContainerLinks, schemas.ContainerMeta
        ](**d)
    else:
        links = {"self": f"{base_url}/metadata/{path_str}"}
        if entry is not None:
            # entry is None when we are pulling just *keys* from the
            # Tree and not values.
            ResourceLinksT = schemas.resource_links_type_by_structure_family[
                entry.structure_family
            ]
            links.update(
                {
                    link: template.format(base_url=base_url, path=path_str)
                    for link, template in FULL_LINKS[entry.structure_family].items()
                }
            )
            structure = asdict(entry.structure())
            if schemas.EntryFields.structure_family in fields:
                attributes["structure_family"] = entry.structure_family
            if entry.structure_family == StructureFamily.sparse:
                shape = structure.get("shape")
                block_template = ",".join(f"{{{index}}}" for index in range(len(shape)))
                links[
                    "block"
                ] = f"{base_url}/array/block/{path_str}?block={block_template}"
            elif entry.structure_family == StructureFamily.array:
                shape = structure.get("shape")
                block_template = ",".join(
                    f"{{index_{index}}}" for index in range(len(shape))
                )
                links[
                    "block"
                ] = f"{base_url}/array/block/{path_str}?block={block_template}"
            elif entry.structure_family == StructureFamily.table:
                links[
                    "partition"
                ] = f"{base_url}/table/partition/{path_str}?partition={{index}}"
            elif entry.structure_family == StructureFamily.awkward:
                links["buffers"] = f"{base_url}/awkward/buffers/{path_str}"
            if schemas.EntryFields.structure in fields:
                attributes["structure"] = structure
        else:
            # We only have entry names, not structure_family, so
            ResourceLinksT = schemas.SelfLinkOnly
        d = {
            "id": path_parts[-1],
            "attributes": schemas.NodeAttributes(**attributes),
        }
        if not omit_links:
            d["links"] = links
        resource = schemas.Resource[
            schemas.NodeAttributes, ResourceLinksT, schemas.EmptyDict
        ](**d)
    return resource


class PatchedResponse(Response):
    "Patch the render method to accept memoryview."

    def render(self, content: Any) -> bytes:
        if isinstance(content, memoryview):
            return content.cast("B")
        return super().render(content)


class PatchedStreamingResponse(StreamingResponse):
    "Patch the stream_response method to accept memoryview."

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for chunk in self.body_iterator:
            # BEGIN ALTERATION
            if not isinstance(chunk, (bytes, memoryview)):
                # END ALTERATION
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})


class NumpySafeJSONResponse(JSONResponse):
    def __init__(self, *args, metrics, **kwargs):
        self.__metrics = metrics
        super().__init__(*args, **kwargs)

    def render(self, content: Any) -> bytes:
        with record_timing(self.__metrics, "pack"):
            return safe_json_dump(content)


def _fallback_msgpack_encoder(obj):
    # If numpy has not been imported yet, then we can be sure that obj
    # is not a numpy object, and we want to avoid triggering a numpy
    # import. (The server does not have a hard numpy dependency.)
    if "numpy" in sys.modules:
        import numpy

        if isinstance(obj, (numpy.generic, numpy.ndarray)):
            if numpy.isscalar(obj):
                return obj.item()
            return obj.tolist()
    if isinstance(obj, uuid.UUID):
        return str(obj)  # hyphen-separated hex per RFC4122
    return obj


def _patch_naive_datetimes(obj):
    """
    If a naive datetime is found, attach local time.

    Msgpack can only serialize datetimes with tzinfo.
    """
    if hasattr(obj, "items"):
        patched_obj = {}
        for k, v in obj.items():
            patched_obj[k] = _patch_naive_datetimes(v)
    elif (not isinstance(obj, str)) and isinstance(obj, collections.abc.Iterable):
        patched_obj = []
        for item in obj:
            patched_obj.append(_patch_naive_datetimes(item))
    elif isinstance(obj, datetime) and obj.tzinfo is None:
        patched_obj = obj.astimezone(_LOCAL_TZINFO)
    else:
        patched_obj = obj
    return patched_obj


class MsgpackResponse(Response):
    media_type = "application/x-msgpack"

    def __init__(self, *args, metrics, **kwargs):
        self.__metrics = metrics
        super().__init__(*args, **kwargs)

    def render(self, content: Any, _reentered=False) -> bytes:
        try:
            with record_timing(self.__metrics, "pack"):
                return msgpack.packb(
                    content, default=_fallback_msgpack_encoder, datetime=True
                )
        except (ValueError, TypeError) as err:
            # msgpack tries to handle all datetimes, but if it
            # received a naive one (tzinfo=None) then it fails.
            # We cannot use the default hook to handle this because
            # it is not called.
            if "can not serialize 'datetime.datetime' object" in str(err) and (
                not _reentered
            ):
                patched_content = _patch_naive_datetimes(content)
                return self.render(patched_content, _reentered=True)
            raise


JSON_MIME_TYPE = "application/json"
MSGPACK_MIME_TYPE = "application/x-msgpack"
# This is a silly time format, but it is the HTTP standard.
HTTP_EXPIRES_HEADER_FORMAT = "%a, %d %b %Y %H:%M:%S GMT"


def resolve_media_type(request):
    media_types = request.headers.get("Accept", JSON_MIME_TYPE).split(", ")
    for media_type in media_types:
        if media_type == "*/*":
            media_type = JSON_MIME_TYPE
            break
        if media_type == MSGPACK_MIME_TYPE:
            break
        if media_type == JSON_MIME_TYPE:
            break
    else:
        # It is commmon in HTTP to fall back on a default representation if
        # none of the requested ones are available. We do not do this for
        # data payloads, but it makes some sense to do it for these metadata
        # messages.
        media_type = JSON_MIME_TYPE
    assert media_type in {JSON_MIME_TYPE, MSGPACK_MIME_TYPE}
    return media_type


def json_or_msgpack(request, content, expires=None, headers=None, status_code=200):
    media_type = resolve_media_type(request)
    with record_timing(request.state.metrics, "tok"):
        etag = md5(str(content).encode()).hexdigest()
    headers = headers or {}
    headers["ETag"] = etag
    if expires is not None:
        headers["Expires"] = expires.strftime(HTTP_EXPIRES_HEADER_FORMAT)
    if request.headers.get("If-None-Match", "") == etag:
        # If the client already has this content, confirm that.
        return Response(status_code=304, headers=headers)
    if media_type == "application/x-msgpack":
        return MsgpackResponse(
            content,
            headers=headers,
            metrics=request.state.metrics,
            status_code=status_code,
        )
    return NumpySafeJSONResponse(
        content, headers=headers, metrics=request.state.metrics, status_code=status_code
    )


class UnsupportedMediaTypes(Exception):
    pass


class NoEntry(KeyError):
    pass


class WrongTypeForRoute(Exception):
    pass


FULL_LINKS = {
    StructureFamily.array: {"full": "{base_url}/array/full/{path}"},
    StructureFamily.awkward: {"full": "{base_url}/awkward/full/{path}"},
    StructureFamily.container: {"full": "{base_url}/container/full/{path}"},
    StructureFamily.table: {"full": "{base_url}/table/full/{path}"},
    StructureFamily.sparse: {"full": "{base_url}/array/full/{path}"},
}


def asdict(dc):
    "Compat for converting dataclass or pydantic.BaseModel to dict."
    if dc is None:
        return None
    try:
        return dataclasses.asdict(dc)
    except TypeError:
        return dict(dc)



import random
import string
from dataclasses import asdict

import dask.array
import numpy
import pandas
import pandas.testing
import pytest
import pytest_asyncio
import sqlalchemy.dialects.postgresql.asyncpg
import sqlalchemy.exc
import tifffile
import xarray

from ..adapters.csv import read_csv
from ..adapters.dataframe import ArrayAdapter
from ..adapters.tiff import TiffAdapter
from ..catalog import in_memory
from ..catalog.adapter import WouldDeleteData
from ..catalog.explain import record_explanations
from ..catalog.register import create_node_safe
from ..catalog.utils import ensure_uri
from ..client import Context, from_context
from ..client.xarray import write_xarray_dataset
from ..queries import Eq, Key
from ..server.app import build_app, build_app_from_config
from ..server.schemas import Asset, DataSource, Management
from ..structures.core import StructureFamily
from .utils import enter_password


@pytest_asyncio.fixture
async def a(adapter):
    "Raw adapter, not to be used within an app becaues it is manually started and stopped."
    await adapter.startup()
    yield adapter
    await adapter.shutdown()


@pytest_asyncio.fixture
async def client(adapter):
    app = build_app(adapter)
    with Context.from_app(app) as context:
        yield from_context(context)


@pytest.mark.asyncio
async def test_nested_node_creation(a):
    await a.create_node(
        key="b",
        metadata={},
        structure_family=StructureFamily.container,
        specs=[],
    )
    b = await a.lookup_adapter(["b"])
    await b.create_node(
        key="c",
        metadata={},
        structure_family=StructureFamily.container,
        specs=[],
    )
    c = await b.lookup_adapter(["c"])
    assert b.segments == ["b"]
    assert c.segments == ["b", "c"]
    assert (await a.keys_range(0, 1)) == ["b"]
    assert (await b.keys_range(0, 1)) == ["c"]
    # smoke test
    await a.items_range(0, 1)
    await b.items_range(0, 1)
    await a.shutdown()


@pytest.mark.asyncio
async def test_sorting(a):
    # Generate lists of letters and numbers, randomly shuffled.
    random_state = random.Random(0)
    ordered_letters = list(string.ascii_lowercase[:10])
    shuffled_letters = list(ordered_letters)
    random_state.shuffle(shuffled_letters)
    shuffled_numbers = [0] * 5 + [1] * 5
    random_state.shuffle(shuffled_numbers)
    assert ordered_letters != shuffled_letters
    assert sorted(shuffled_letters) != shuffled_letters

    for letter, number in zip(shuffled_letters, shuffled_numbers):
        await a.create_node(
            key=letter,
            metadata={"letter": letter, "number": number},
            structure_family=StructureFamily.container,
            specs=[],
        )

    # Default sorting is _not_ ordered.
    default_key_order = await a.keys_range(0, 10)
    assert default_key_order != ordered_letters
    # Sorting by ("", -1) gives reversed default order.
    reversed_default_key_order = await a.sort([("", -1)]).keys_range(0, 10)
    assert reversed_default_key_order == list(reversed(default_key_order))

    # Sort by key.
    await a.sort([("id", 1)]).keys_range(0, 10) == ordered_letters
    # Test again, with items_range.
    [k for k, v in await a.sort([("id", 1)]).items_range(0, 10)] == ordered_letters

    # Sort by letter metadata.
    # Use explicit 'metadata.{key}' namespace.
    assert (await a.sort([("metadata.letter", 1)]).keys_range(0, 10)) == ordered_letters

    # Sort by letter metadata.
    # Use implicit '{key}' (more convenient, and necessary for back-compat).
    assert (await a.sort([("letter", 1)]).keys_range(0, 10)) == ordered_letters

    # Sort by number and then by letter.
    # Use explicit 'metadata.{key}' namespace.
    items = await a.sort([("metadata.number", 1), ("metadata.letter", 1)]).items_range(
        0, 10
    )
    numbers = [v.metadata()["number"] for k, v in items]
    letters = [v.metadata()["letter"] for k, v in items]
    keys = [k for k, v in items]
    # Numbers are sorted.
    numbers = sorted(numbers)
    # Within each block of numbers, keys and letters are sorted.
    assert sorted(keys[:5]) == keys[:5] == letters[:5]
    assert sorted(keys[5:]) == keys[5:] == letters[5:]


@pytest.mark.asyncio
async def test_search(a):
    for letter, number in zip(string.ascii_lowercase[:5], range(5)):
        await a.create_node(
            key=letter,
            metadata={"letter": letter, "number": number, "x": {"y": {"z": letter}}},
            structure_family=StructureFamily.container,
            specs=[],
        )
    assert "c" in await a.keys_range(0, 5)
    assert await a.search(Eq("letter", "c")).keys_range(0, 5) == ["c"]
    assert await a.search(Eq("number", 2)).keys_range(0, 5) == ["c"]

    # Looking up "d" inside search results should find nothing when
    # "d" is filtered out by a search query first.
    assert await a.lookup_adapter(["d"]) is not None
    assert await a.search(Eq("letter", "c")).lookup_adapter(["d"]) is None

    # Search on nested key.
    assert await a.search(Eq("x.y.z", "c")).keys_range(0, 5) == ["c"]

    # Created nested nodes and search on them.
    d = await a.lookup_adapter(["d"])
    for letter, number in zip(string.ascii_lowercase[:5], range(10, 15)):
        await d.create_node(
            key=letter,
            metadata={"letter": letter, "number": number},
            structure_family=StructureFamily.container,
            specs=[],
        )
    assert await d.search(Eq("letter", "c")).keys_range(0, 5) == ["c"]
    assert await d.search(Eq("number", 12)).keys_range(0, 5) == ["c"]


@pytest.mark.asyncio
async def test_metadata_index_is_used(example_data_adapter):
    a = example_data_adapter  # for succinctness below
    # Check that an index (specifically the 'top_level_metadata' index) is used
    # by inspecting the content of an 'EXPLAIN ...' query. The exact content
    # is intended for humans and is not an API, but we can coarsely check
    # that the index of interest is mentioned.
    await a.startup()
    with record_explanations() as e:
        results = await a.search(Key("number_as_string") == "3").keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("number") == 3).keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("bool") == False).keys_range(0, 5)  # noqa: #712
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("nested.number_as_string") == "3").keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("nested.number") == 3).keys_range(0, 5)
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    with record_explanations() as e:
        results = await a.search(Key("nested.bool") == False).keys_range(  # noqa: #712
            0, 5
        )
        assert len(results) == 1
        assert "top_level_metadata" in str(e)
    await a.shutdown()


@pytest.mark.asyncio
async def test_write_array_external(a, tmpdir):
    arr = numpy.ones((5, 3))
    filepath = str(tmpdir / "file.tiff")
    data_uri = ensure_uri(filepath)
    tifffile.imwrite(filepath, arr)
    ad = TiffAdapter(data_uri)
    structure = asdict(ad.structure())
    await a.create_node(
        key="x",
        structure_family="array",
        metadata={},
        data_sources=[
            DataSource(
                mimetype="image/tiff",
                structure=structure,
                parameters={},
                management="external",
                assets=[
                    Asset(
                        parameter="data_uri",
                        num=None,
                        data_uri=str(data_uri),
                        is_directory=False,
                    )
                ],
            )
        ],
    )
    x = await a.lookup_adapter(["x"])
    assert numpy.array_equal(await x.read(), arr)


@pytest.mark.asyncio
async def test_write_dataframe_external_direct(a, tmpdir):
    df = pandas.DataFrame(numpy.ones((5, 3)), columns=list("abc"))
    filepath = str(tmpdir / "file.csv")
    data_uri = ensure_uri(filepath)
    df.to_csv(filepath, index=False)
    dfa = read_csv(data_uri)
    structure = asdict(dfa.structure())
    await a.create_node(
        key="x",
        structure_family=StructureFamily.table,
        metadata={},
        data_sources=[
            DataSource(
                mimetype="text/csv",
                structure=structure,
                parameters={},
                management="external",
                assets=[
                    Asset(
                        parameter="data_uri",
                        num=None,
                        data_uri=data_uri,
                        is_directory=False,
                    )
                ],
            )
        ],
    )
    x = await a.lookup_adapter(["x"])
    pandas.testing.assert_frame_equal(await x.read(), df)


@pytest.mark.asyncio
async def test_write_array_internal_direct(a, tmpdir):
    arr = numpy.ones((5, 3))
    ad = ArrayAdapter.from_array(arr)
    structure = asdict(ad.structure())
    await a.create_node(
        key="x",
        structure_family="array",
        metadata={},
        data_sources=[
            DataSource(
                structure=structure,
                management="writable",
            )
        ],
    )
    x = await a.lookup_adapter(["x"])
    await x.write(arr)
    assert numpy.array_equal(await x.read(), arr)


def test_write_array_internal_via_client(client):
    expected = numpy.array([1, 3, 7])
    x = client.write_array(expected)
    actual = x.read()
    assert numpy.array_equal(actual, expected)

    y = client.write_array(dask.array.from_array(expected, chunks=((1, 1, 1),)))
    actual = y.read()
    assert numpy.array_equal(actual, expected)


def test_write_dataframe_internal_via_client(client):
    expected = pandas.DataFrame(numpy.ones((5, 3)), columns=list("abc"))
    x = client.write_dataframe(expected)
    actual = x.read()
    pandas.testing.assert_frame_equal(actual, expected)

    # y = client.write_array(dask.array.from_array(expected, chunks=((1, 1, 1),)))
    # actual = y.read()
    # assert numpy.array_equal(actual, expected)
    # pandas.testing.assert_frame_equal(actual, expected)


def test_write_xarray_dataset(client):
    ds = xarray.Dataset(
        {"temp": (["time"], [101, 102, 103])},
        coords={"time": (["time"], [1, 2, 3])},
    )
    dsc = write_xarray_dataset(client, ds, key="test_xarray_dataset")
    assert set(dsc) == {"temp", "time"}
    # smoke test
    dsc["temp"][:]
    dsc["time"][:]
    dsc.read()


@pytest.mark.asyncio
async def test_delete_tree(tmpdir):
    # Do not use client fixture here.
    # The Context must be opened inside the test or we run into
    # event loop crossing issues with the Postgres test.
    tree = in_memory(writable_storage=tmpdir)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        a = client.create_container("a")
        b = a.create_container("b")
        b.write_array([1, 2, 3])
        b.write_array([4, 5, 6])
        c = b.create_container("c")
        d = c.create_container("d")
        d.write_array([7, 8, 9])

        nodes_before_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_before_delete) == 7
        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 3
        assets_before_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_before_delete) == 3

        with pytest.raises(WouldDeleteData):
            await tree.delete_tree()  # external_only=True by default
        with pytest.raises(WouldDeleteData):
            await tree.delete_tree(external_only=True)
        await tree.delete_tree(external_only=False)

        nodes_after_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_after_delete) == 0
        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 0
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 0


@pytest.mark.asyncio
async def test_access_control(tmpdir):
    config = {
        "authentication": {
            "allow_anonymous_access": True,
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {
                            "alice": "secret1",
                            "bob": "secret2",
                            "admin": "admin",
                        }
                    },
                }
            ],
        },
        "database": {
            "uri": "sqlite+aiosqlite://",  # in-memory
        },
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {
                    "uri": f"sqlite+aiosqlite:///{tmpdir}/catalog.db",
                    "writable_storage": str(tmpdir / "data"),
                    "init_if_not_exists": True,
                },
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": ["outer_x"],
                            "bob": ["outer_y"],
                        },
                        "admins": ["admin"],
                        "public": ["outer_z"],
                    },
                },
            },
        ],
    }

    app = build_app_from_config(config)
    with Context.from_app(app) as context:
        with enter_password("admin"):
            admin_client = from_context(context, username="admin")
            for key in ["outer_x", "outer_y", "outer_z"]:
                container = admin_client.create_container(key)
                container.write_array([1, 2, 3], key="inner")
            admin_client.logout()
        with enter_password("secret1"):
            alice_client = from_context(context, username="alice")
            alice_client["outer_x"]["inner"].read()
            with pytest.raises(KeyError):
                alice_client["outer_y"]
            alice_client.logout()
        public_client = from_context(context)
        public_client["outer_z"]["inner"].read()
        with pytest.raises(KeyError):
            public_client["outer_x"]


@pytest.mark.parametrize(
    "assets",
    [
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
        ],
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
        ],
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=None,
            ),
        ],
        [
            Asset(
                data_uri="file://localhost/test1",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
            Asset(
                data_uri="file://localhost/test2",
                is_directory=False,
                parameter="filepath",
                num=1,
            ),
        ],
    ],
    ids=[
        "null-then-int",
        "int-then-null",
        "duplicate-null",
        "duplicate-int",
    ],
)
@pytest.mark.asyncio
async def test_constraints_on_parameter_and_num(a, assets):
    "Test constraints enforced by database on 'parameter' and 'num'."
    arr_adapter = ArrayAdapter.from_array([1, 2, 3])
    with pytest.raises(
        (
            sqlalchemy.exc.IntegrityError,  # SQLite
            sqlalchemy.exc.DBAPIError,  # PostgreSQL
        )
    ):
        await create_node_safe(
            a,
            key="test",
            structure_family=arr_adapter.structure_family,
            metadata=dict(arr_adapter.metadata()),
            specs=arr_adapter.specs,
            data_sources=[
                DataSource(
                    mimetype="application/x-test",
                    structure=asdict(arr_adapter.structure()),
                    parameters={},
                    management=Management.external,
                    assets=assets,
                )
            ],
        )



import numpy
import pytest
import sparse

from ..adapters.mapping import MapAdapter
from ..adapters.sparse import COOAdapter
from ..client import Context, from_context
from ..server.app import build_app

N, M = 3, 5
state = numpy.random.RandomState(0)
a = state.random((N * 2, M * 2))
a[a < 0.5] = 0  # Fill half of the array with zeros.
s = sparse.COO(a)
blocks = {}
for i in range(2):
    for j in range(2):
        chunk = s[i * N : (1 + i) * N, j * M : (1 + j) * M]  # noqa: E203
        # Below we test the COOAdapter.from_global_ref constructor, so
        # coordinates should be in the reference frame of the whole array.
        coords = chunk.coords + [[i * N], [j * M]]
        blocks[i, j] = coords, chunk.data
chunks = ((N, N), (M, M))
dims = ["x", "y"]
mapping = {
    "single_chunk": COOAdapter.from_coo(sparse.COO(a), dims=dims),
    "multi_chunk": COOAdapter.from_global_ref(
        blocks=blocks, shape=(2 * N, 2 * M), chunks=chunks, dims=dims
    ),
}
tree = MapAdapter(mapping)


@pytest.fixture(scope="module")
def client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


def test_sparse_single_chunk(client):
    sc = client["single_chunk"]
    actual_via_slice = sc[:]
    actual_via_read = sc.read()
    actual_via_todense = sc.todense()
    actual_via_read_block = sc.read_block((0, 0))
    assert numpy.array_equal(actual_via_slice.todense(), actual_via_read.todense())
    assert numpy.array_equal(
        actual_via_slice.todense(), actual_via_read_block.todense()
    )
    assert numpy.array_equal(actual_via_slice.todense(), a)
    assert numpy.array_equal(actual_via_todense, a)
    assert sc.shape == a.shape
    assert sc.ndim == a.ndim
    assert sc.chunks == tuple((i,) for i in a.shape)
    assert sc.dims == dims


def test_sparse_multi_chunk(client):
    sc = client["multi_chunk"]
    actual_via_slice = sc[:]
    actual_via_read = sc.read()
    actual_via_todense = sc.todense()
    assert numpy.array_equal(actual_via_slice.todense(), actual_via_read.todense())
    assert numpy.array_equal(actual_via_slice.todense(), a)
    assert numpy.array_equal(actual_via_todense, a)
    assert sc.shape == a.shape
    assert sc.ndim == a.ndim
    assert sc.chunks == chunks
    assert sc.dims == dims



import pytest

from ..utils import CachingMap, OneShotCachedMap


def test_one_shot_cached_map():
    OneShotCachedMap({})  # empty is allowed

    counter = 0

    def f():
        nonlocal counter
        counter += 1
        return 5

    d = OneShotCachedMap({"a": f})
    # These operations do not cause f to be called.
    len(d)
    "a" in d
    list(d)
    assert counter == 0

    assert d["a"] == 5  # f is called here
    assert counter == 1
    assert d["a"] == 5  # but not again
    assert counter == 1

    with pytest.raises(TypeError):
        del d["a"]
    with pytest.raises(TypeError):
        d["a"] = 5


def test_caching_map():
    CachingMap({}, {})  # empty is allowed

    counter = 0

    def f():
        nonlocal counter
        counter += 1
        return 5

    cache = {}
    mapping = {"a": f}
    d = CachingMap(mapping, cache)

    # These operations do not cause f to be called.
    len(d)
    "a" in d
    list(d)
    assert counter == 0

    assert d["a"] == 5  # f is called here
    assert counter == 1
    assert d["a"] == 5  # but not again
    assert counter == 1
    cache.clear()  # underlying cache evicts
    assert d["a"] == 5  # f is called here again
    assert counter == 2

    # Normal mutation is not allowed.
    with pytest.raises(TypeError):
        del d["a"]
    with pytest.raises(TypeError):
        d["a"] = 5

    # Special mutation methods
    assert "a" in cache
    assert "a" in mapping
    assert "a" in d
    d.evict("a")
    assert "a" not in cache
    assert "a" in mapping
    assert "a" in d
    d.evict("a")  # idempotent
    assert "a" not in cache
    assert "a" in mapping
    assert "a" in d
    assert d["a"] == 5  # f is called here again
    assert counter == 3
    d.remove("a")
    assert "a" not in cache
    assert "a" not in mapping
    assert "a" not in d
    with pytest.raises(KeyError):
        d.remove("a")
    with pytest.raises(KeyError):
        d.remove("never existed")
    d.set("a", f)
    assert d["a"] == 5  # f is called here again
    assert counter == 4
    d.discard("a")
    assert "a" not in cache
    assert "a" not in mapping
    assert "a" not in d
    d.discard("a")  # idempotent
    d.discard("never existed")



from pathlib import Path

import pytest

from ..catalog.register import register
from ..client import Context, from_context
from ..examples.xdi import data
from ..server.app import build_app_from_config


@pytest.mark.asyncio
async def test_xdi_round_trip(tmpdir):
    """
    Steps:

    1. Parse XDI string into DataFrame + dict.
    2. Export as XDI to client.
    3. Read exported XDI into DataFrame + dict.

    Compare result of (3) to result of (1).
    """
    # Write example data file.
    Path(tmpdir / "files").mkdir()
    with open(tmpdir / "files" / "example.xdi", "w") as file:
        file.write(data)
    config = {
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {
                    "uri": tmpdir / "catalog.db",
                    "readable_storage": [tmpdir / "files"],
                    "init_if_not_exists": True,
                    "adapters_by_mimetype": {
                        "application/x-xdi": "tiled.examples.xdi:read_xdi"
                    },
                },
            }
        ],
        "media_types": {"xdi": {"application/x-xdi": "tiled.examples.xdi:write_xdi"}},
        "file_extensions": {"xdi": "application/x-xdi"},
    }
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        tree = context.http_client.app.state.root_tree
        await register(
            tree,
            tmpdir / "files",
            adapters_by_mimetype={"application/x-xdi": "tiled.examples.xdi:read_xdi"},
            mimetypes_by_file_ext={".xdi": "application/x-xdi"},
        )
        client["example"].export(str(tmpdir / "exported.xdi"))
        actual = Path(tmpdir / "exported.xdi").read_text()
        actual
        # XDI uses a two-space spacing that pandas.to_csv does not support.
        # Need thought to get this exactly right.
        # assert actual == data



from typing import Optional

try:
    import typer
except Exception as err:
    raise Exception(
        """
from ._admin import admin_app
from ._api_key import api_key_app
from ._profile import profile_app
from ._serve import serve_app

You are trying to the run the tiled commandline tool but you do not have the
necessary dependencies. It looks like tiled has been installed with
bare-minimum dependencies, possibly via

    pip install tiled

Instead, try:

    pip install tiled[all]  # Note: on a Mac, you may need quotes like 'tiled[all]'.

which installs *everything* you might want. For other options, see:

    https://blueskyproject.io/tiled/tutorials/installation.html
"""
    ) from err

from ..utils import UNSET

cli_app = typer.Typer()

from ._admin import admin_app  # noqa: E402
from ._api_key import api_key_app  # noqa: E402
from ._catalog import catalog_app  # noqa: E402
from ._profile import profile_app  # noqa: E402
from ._serve import serve_app  # noqa: E402
from ._utils import get_context, get_profile  # noqa E402

cli_app.add_typer(
    catalog_app, name="catalog", help="Manage a catalog of data to be served by Tiled."
)
cli_app.add_typer(serve_app, name="serve", help="Launch a Tiled server.")
cli_app.add_typer(
    profile_app, name="profile", help="Examine Tiled 'profiles' (client-side config)."
)
cli_app.add_typer(
    api_key_app, name="api_key", help="Create, list, and revoke API keys."
)
cli_app.add_typer(
    admin_app,
    name="admin",
    help="Administrative utilities for managing large deployments.",
)


@cli_app.command("login")
def login(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    set_default: bool = typer.Option(
        True, help="Use this identity as the default for this API."
    ),
    show_secret_tokens: bool = typer.Option(
        False, "--show-secret-tokens", help="Show secret tokens after successful login."
    ),
):
    """
    Log in to an authenticated Tiled server.
    """
    from ..client.context import Context

    profile_name, profile_content = get_profile(profile)
    options = {"verify": profile_content.get("verify", True)}
    context, _ = Context.from_any_uri(profile_content["uri"], **options)
    # Override sticky 'default_identity'.
    # Always prompt user to specify who they want to log in as.
    context.authenticate(username=None, provider=None, set_default=True)
    if show_secret_tokens:
        import json

        typer.echo(json.dumps(dict(context.tokens), indent=4))


@cli_app.command("logout")
def logout(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    username: Optional[str] = typer.Option(None),
    provider: Optional[str] = typer.Option(None),
):
    """
    Log out.
    """
    from ..client.context import Context

    profile_name, profile_content = get_profile(profile)
    context, _ = Context.from_any_uri(
        profile_content["uri"], verify=profile_content.get("verify", True)
    )
    if username is None:
        username = UNSET
    if provider is None:
        provider = UNSET
    context.authenticate(username=username, provider=provider, set_default=False)
    context.logout()


@cli_app.command("tree")
def tree(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    max_lines: int = typer.Argument(20, help="Max lines to show."),
):
    """
    Show the names of entries in a Tree.

    This is similar to the UNIX utility `tree` for listing nested directories.
    """
    from ..client.constructors import from_context
    from ..utils import gen_tree

    context = get_context(profile)
    client = from_context(context)
    for counter, line in enumerate(gen_tree(client), start=1):
        if (max_lines is not None) and (counter > max_lines):
            print(
                f"Output truncated at {max_lines} lines. "
                "Use `tiled tree <N>` to see <N> lines."
            )
            break
        print(line)


main = cli_app


if __name__ == "__main__":
    main()

# This object is used by the auto-generated documentation.
typer_click_object = typer.main.get_command(cli_app)



from pathlib import Path

import httpx
import pytest
import yaml

from ..adapters.mapping import MapAdapter
from ..client import Context, from_context, from_profile, record_history
from ..config import ConfigError
from ..profiles import load_profiles, paths
from ..queries import Key
from ..server.app import build_app
from .utils import fail_with_status_code

tree = MapAdapter({})


def test_configurable_timeout():
    with Context.from_app(build_app(tree), timeout=httpx.Timeout(17)) as context:
        assert context.http_client.timeout.connect == 17
        assert context.http_client.timeout.read == 17


def test_client_version_check():
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        # Too-old user agent should generate a 400.
        context.http_client.headers["user-agent"] = "python-tiled/0.1.0a77"
        with fail_with_status_code(400):
            list(client)

        # Gibberish user agent should generate a 400.
        context.http_client.headers["user-agent"] = "python-tiled/gibberish"
        with fail_with_status_code(400):
            list(client)


def test_direct(tmpdir):
    profile_content = {
        "test": {
            "structure_clients": "dask",
            "direct": {
                "trees": [
                    {"path": "/", "tree": "tiled.examples.generated_minimal:tree"}
                ]
            },
        }
    }
    with open(tmpdir / "example.yml", "w") as file:
        file.write(yaml.dump(profile_content))
    profile_dir = Path(tmpdir)
    try:
        paths.append(profile_dir)
        load_profiles.cache_clear()
        from_profile("test")
    finally:
        paths.remove(profile_dir)


def test_direct_config_error(tmpdir):
    profile_content = {
        "test": {
            "direct": {
                # Intentional config mistake!
                # Value of trees must be a list.
                "trees": {"path": "/", "tree": "tiled.examples.generated_minimal:tree"}
            }
        }
    }
    with open(tmpdir / "example.yml", "w") as file:
        file.write(yaml.dump(profile_content))
    profile_dir = Path(tmpdir)
    try:
        paths.append(profile_dir)
        load_profiles.cache_clear()
        with pytest.raises(ConfigError):
            from_profile("test")
    finally:
        paths.remove(profile_dir)


def test_jump_down_tree():
    tree = MapAdapter({}, metadata={"number": 1})
    for number, letter in enumerate(list("abcde"), start=2):
        tree = MapAdapter({letter: tree}, metadata={"number": number})
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert (
        client["e"]["d"]["c"]["b"]["a"].metadata["number"]
        == client["e", "d", "c", "b", "a"].metadata["number"]
        == 1
    )
    assert (
        client["e"]["d"]["c"]["b"].metadata["number"]
        == client["e", "d", "c", "b"].metadata["number"]
        == 2
    )
    assert (
        client["e"]["d"]["c"].metadata["number"]
        == client["e", "d", "c"].metadata["number"]
        == 3
    )
    assert (
        client["e"]["d"].metadata["number"] == client["e", "d"].metadata["number"] == 4
    )

    assert client["e"]["d", "c", "b"]["a"].metadata["number"] == 1
    assert client["e"]["d", "c", "b", "a"].metadata["number"] == 1
    assert client["e", "d", "c", "b"]["a"].metadata["number"] == 1
    assert (
        client.search(Key("number") == 5)["e", "d", "c", "b", "a"].metadata["number"]
        == 1
    )
    assert (
        client["e"].search(Key("number") == 4)["d", "c", "b", "a"].metadata["number"]
        == 1
    )

    # Check that a reasonable KeyError is raised.
    # Notice that we do not binary search to find _exactly_ where the problem is.
    with pytest.raises(KeyError) as exc_info:
        client["e", "d", "c", "b"]["X"]
    assert exc_info.value.args[0] == "X"
    with pytest.raises(KeyError) as exc_info:
        client["e", "d", "c", "b", "X"]
    assert exc_info.value.args[0] == ("e", "d", "c", "b", "X")
    with pytest.raises(KeyError) as exc_info:
        client["e", "d", "X", "b", "a"]
    assert exc_info.value.args[0] == ("e", "d", "X", "b", "a")

    # Check that jumping raises if a key along the path is not in the search
    # resuts.
    with pytest.raises(KeyError) as exc_info:
        client.search(Key("number") == 4)["e"]
    assert exc_info.value.args[0] == "e"
    with pytest.raises(KeyError) as exc_info:
        client.search(Key("number") == 4)["e", "d", "c", "b", "a"]
    assert exc_info.value.args[0] == "e"
    with pytest.raises(KeyError) as exc_info:
        client["e"].search(Key("number") == 3)["d"]
    assert exc_info.value.args[0] == "d"
    with pytest.raises(KeyError) as exc_info:
        client["e"].search(Key("number") == 3)["d", "c", "b", "a"]
    assert exc_info.value.args[0] == "d"

    with record_history() as h:
        client["e", "d", "c", "b", "a"]
    assert len(h.requests) == 1

    with record_history() as h:
        client["e"]["d"]["c"]["b"]["a"]
    assert len(h.requests) == 5
